import logging
from cis.data_io import hdf as hdf
from cis.data_io.Coord import CoordList, Coord
from cis.data_io.products import AProduct
from cis.data_io.ungridded_data import UngriddedCoordinates, UngriddedData


def _get_MODIS_SDS_data(sds):
    """
    Reads raw data from an SD instance.

    :param sds: The specific sds instance to read
    :return: A numpy array containing the raw data with missing data is replaced by NaN.
    """
    from cis.utils import create_masked_array_for_missing_data
    import numpy as np

    data = sds.get()
    attributes = sds.attributes()

    # Apply Fill Value
    missing_value = attributes.get('_FillValue', None)
    if missing_value is not None:
        data = create_masked_array_for_missing_data(data, missing_value)

    # Check for valid_range
    valid_range = attributes.get('valid_range', None)
    if valid_range is not None:
        logging.debug("Masking all values {} > v > {}.".format(*valid_range))
        data = np.ma.masked_outside(data, *valid_range)

    # Offsets and scaling.
    add_offset = attributes.get('add_offset', 0.0)
    scale_factor = attributes.get('scale_factor', 1.0)
    data = _apply_scaling_factor_MODIS(data, scale_factor, add_offset)

    return data


def _apply_scaling_factor_MODIS(data, scale_factor, offset):
    """
    Apply scaling factor (applicable to MODIS data) of the form:
    ``data = (data - offset) * scale_factor``

    Ref:
    MODIS Atmosphere L3 Gridded Product Algorithm Theoretical Basis Document,
    MODIS Algorithm Theoretical Basis Document No. ATBD-MOD-30 for
    Level-3 Global Gridded Atmosphere Products (08_D3, 08_E3, 08_M3)
    by PAUL A. HUBANKS, MICHAEL D. KING, STEVEN PLATNICK, AND ROBERT PINCUS
    (Collection 005 Version 1.1, 4 December 2008)

    :param data: A numpy array like object
    :param scale_factor:
    :param offset:
    :return: Scaled data
    """
    logging.debug("Applying 'science_data = (packed_data - {offset}) * {scale}' "
                  "transformation to data.".format(scale=scale_factor, offset=offset))
    return (data - offset) * scale_factor


class MODIS_L3(AProduct):
    """
    Data product for MODIS Level 3 data
    """

    def _parse_datetime(self, metadata_dict, keyword):
        import re
        res = ""
        for s in metadata_dict.values():
            i_start = s.find(keyword)
            ssub = s[i_start:len(s)]
            i_end = ssub.find("END_OBJECT")
            ssubsub = s[i_start:i_start + i_end]
            matches = re.findall('".*"', ssubsub)
            if len(matches) > 0:
                res = matches[0].replace('\"', '')
                if res is not "":
                    break
        return res

    def _get_start_date(self, filename):
        from cis.parse_datetime import parse_datetimestr_to_std_time
        metadata_dict = hdf.get_hdf4_file_metadata(filename)
        date = self._parse_datetime(metadata_dict, 'RANGEBEGINNINGDATE')
        time = self._parse_datetime(metadata_dict, 'RANGEBEGINNINGTIME')
        datetime_str = date + " " + time
        return parse_datetimestr_to_std_time(datetime_str)

    def _get_end_date(self, filename):
        from cis.parse_datetime import parse_datetimestr_to_std_time
        metadata_dict = hdf.get_hdf4_file_metadata(filename)
        date = self._parse_datetime(metadata_dict, 'RANGEENDINGDATE')
        time = self._parse_datetime(metadata_dict, 'RANGEENDINGTIME')
        datetime_str = date + " " + time
        return parse_datetimestr_to_std_time(datetime_str)

    def get_file_signature(self):
        product_names = ['MYD08_D3', 'MOD08_D3', 'MYD08_M3', 'MOD08_M3', "MOD08_E3"]
        regex_list = [r'.*' + product + '.*\.hdf' for product in product_names]
        return regex_list

    def get_variable_names(self, filenames, data_type=None):
        try:
            from pyhdf.SD import SD
        except ImportError:
            raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

        variables = set([])
        for filename in filenames:
            sd = SD(filename)
            for var_name, var_info in sd.datasets().items():
                # Check that the dimensions are correct
                if var_info[0] == ('YDim:mod08', 'XDim:mod08'):
                    variables.add(var_name)

        return variables

    def _create_cube(self, filenames, variable):
        import numpy as np
        from cis.data_io.hdf import _read_hdf4
        from iris.cube import Cube, CubeList
        from iris.coords import DimCoord, AuxCoord
        from cis.time_util import calculate_mid_time, cis_standard_time_unit
        from cis.data_io.hdf_sd import get_metadata
        from cf_units import Unit

        variables = ['XDim', 'YDim', variable]
        logging.info("Listing coordinates: " + str(variables))

        cube_list = CubeList()
        # Read each file individually, let Iris do the merging at the end.
        for f in filenames:
            sdata, vdata = _read_hdf4(f, variables)

            lat_coord = DimCoord(_get_MODIS_SDS_data(sdata['YDim']), standard_name='latitude', units='degrees')
            lon_coord = DimCoord(_get_MODIS_SDS_data(sdata['XDim']), standard_name='longitude', units='degrees')

            # create time coordinate using the midpoint of the time delta between the start date and the end date
            start_datetime = self._get_start_date(f)
            end_datetime = self._get_end_date(f)
            mid_datetime = calculate_mid_time(start_datetime, end_datetime)
            logging.debug("Using {} as datetime for file {}".format(mid_datetime, f))
            time_coord = AuxCoord(mid_datetime, standard_name='time', units=cis_standard_time_unit,
                                  bounds=[start_datetime, end_datetime])

            var = sdata[variable]
            metadata = get_metadata(var)

            try:
                units = Unit(metadata.units)
            except ValueError:
                logging.warning("Unable to parse units '{}' in {} for {}.".format(metadata.units, f, variable))
                units = None

            cube = Cube(_get_MODIS_SDS_data(sdata[variable]),
                        dim_coords_and_dims=[(lon_coord, 1), (lat_coord, 0)],
                        aux_coords_and_dims=[(time_coord, None)],
                        var_name=metadata._name, long_name=metadata.long_name, units=units)

            cube_list.append(cube)

        # Merge the cube list across the scalar time coordinates before returning a single cube.
        return cube_list.merge_cube()

    def create_coords(self, filenames, variable=None):
        """Reads the coordinates on which a variable depends.
        Note: This calls create_data_object because the coordinates are returned as a Cube.
        :param filenames: list of names of files from which to read coordinates
        :param variable: name of variable for which the coordinates are required
        :return: iris.cube.Cube
        """
        if variable is None:
            variable_names = self.get_variable_names(filenames)
            variable_name = str(variable_names.pop())
            logging.debug("Reading an IRIS Cube for the coordinates based on the variable %s" % variable_names)
        else:
            variable_name = variable

        return self.create_data_object(filenames, variable_name)

    def create_data_object(self, filenames, variable):
        from cis.data_io.gridded_data import make_from_cube
        logging.debug("Creating data object for variable " + variable)

        cube = self._create_cube(filenames, variable)
        return make_from_cube(cube)

    def get_file_format(self, filename):
        return "HDF4/ModisL3"


class MODIS_L2(AProduct):
    modis_scaling = ["1km", "5km", "10km"]

    def get_file_signature(self):
        product_names = ['MYD06_L2', 'MOD06_L2', 'MYD04_L2', 'MOD04_L2']
        regex_list = [r'.*' + product + '.*\.hdf' for product in product_names]
        return regex_list

    def get_variable_names(self, filenames, data_type=None):
        import pyhdf.SD

        # Determine the valid shape for variables
        sd = pyhdf.SD.SD(filenames[0])
        datasets = sd.datasets()
        valid_shape = datasets['Latitude'][1]  # Assumes that latitude shape == longitude shape (it should)

        variables = set([])
        for filename in filenames:
            sd = pyhdf.SD.SD(filename)
            for var_name, var_info in sd.datasets().items():
                if var_info[1] == valid_shape:
                    variables.add(var_name)

        return variables

    def __get_data_scale(self, filename, variable):
        from cis.exceptions import InvalidVariableError
        from pyhdf.SD import SD

        try:
            meta = SD(filename).datasets()[variable][0][0]
        except KeyError:
            raise InvalidVariableError("Variable " + variable + " not found")

        for scaling in self.modis_scaling:
            if scaling in meta:
                return scaling
        return None

    def __field_interpolate(self, data, factor=5):
        """
        Interpolates the given 2D field by the factor,
        edge pixels are defined by the ones in the centre,
        odd factords only!
        """
        import numpy as np

        logging.debug("Performing interpolation...")

        output = np.zeros((factor * data.shape[0], factor * data.shape[1])) * np.nan
        output[int(factor / 2)::factor, int(factor / 2)::factor] = data
        for i in range(1, factor + 1):
            output[(int(factor / 2) + i):(-1 * factor / 2 + 1):factor, :] = i * (
                (output[int(factor / 2) + factor::factor, :] - output[int(factor / 2):(-1 * factor):factor, :]) /
                float(factor)) + output[int(factor / 2):(-1 * factor):factor, :]
        for i in range(1, factor + 1):
            output[:, (int(factor / 2) + i):(-1 * factor / 2 + 1):factor] = i * (
                (output[:, int(factor / 2) + factor::factor] - output[:, int(factor / 2):(-1 * factor):factor]) /
                float(factor)) + output[:, int(factor / 2):(-1 * factor):factor]
        return output

    def _create_coord_list(self, filenames, variable=None):
        import datetime as dt

        variables = ['Latitude', 'Longitude', 'Scan_Start_Time']
        logging.info("Listing coordinates: " + str(variables))

        sdata, vdata = hdf.read(filenames, variables)

        apply_interpolation = False
        if variable is not None:
            scale = self.__get_data_scale(filenames[0], variable)
            apply_interpolation = True if scale is "1km" else False

        lat = sdata['Latitude']
        sd_lat = hdf.read_data(lat, _get_MODIS_SDS_data)
        lat_data = self.__field_interpolate(sd_lat) if apply_interpolation else sd_lat
        lat_metadata = hdf.read_metadata(lat, "SD")
        lat_coord = Coord(lat_data, lat_metadata, 'Y')

        lon = sdata['Longitude']
        if apply_interpolation:
            lon_data = self.__field_interpolate(hdf.read_data(lon, _get_MODIS_SDS_data))
        else:
            lon_data = hdf.read_data(lon, _get_MODIS_SDS_data)

        lon_metadata = hdf.read_metadata(lon, "SD")
        lon_coord = Coord(lon_data, lon_metadata, 'X')

        time = sdata['Scan_Start_Time']
        time_metadata = hdf.read_metadata(time, "SD")
        # Ensure the standard name is set
        time_metadata.standard_name = 'time'
        time_coord = Coord(time, time_metadata, "T", _get_MODIS_SDS_data)
        time_coord.convert_TAI_time_to_std_time(dt.datetime(1993, 1, 1, 0, 0, 0))

        return CoordList([lat_coord, lon_coord, time_coord])

    def create_coords(self, filenames, variable=None):
        return UngriddedCoordinates(self._create_coord_list(filenames))

    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self._create_coord_list(filenames, variable)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        return UngriddedData(var, metadata, coords, _get_MODIS_SDS_data)


    def get_file_format(self, filenames):
        """
        Get the file format
        :param filenames: the filenames of the file
        :return: file format
        """

        return "HDF4/ModisL2"



import logging
from cis.data_io import hdf as hdf
from cis.data_io.Coord import Coord, CoordList
from cis.data_io.hdf_vd import get_data, VDS
from cis.data_io.products import AProduct
from cis.data_io.ungridded_data import Metadata, UngriddedCoordinates, UngriddedData
import cis.utils as utils

MIXED_RESOLUTION_VARIABLES = ['Atmospheric_Volume_Description', 'CAD_Score',
                              'Extinction_QC_Flag_1064', 'Extinction_QC_Flag_532']


class abstract_Caliop(AProduct):

    include_pressure = True

    def get_file_signature(self):
        '''
        To be implemented by subclcass
        :return:
        '''
        return []

    def get_variable_names(self, filenames, data_type=None):
        try:
            from pyhdf.SD import SD
        except ImportError:
            raise ImportError("HDF support was not installed, please reinstall with pyhdf to read HDF files.")

        variables = set([])

        # Determine the valid shape for variables
        sd = SD(filenames[0])
        datasets = sd.datasets()
        len_x = datasets['Latitude'][1][0]  # Assumes that latitude shape == longitude shape (it should)
        alt_data = get_data(VDS(filenames[0], "Lidar_Data_Altitudes"), True)
        len_y = alt_data.shape[0]
        valid_shape = (len_x, len_y)

        for filename in filenames:
            sd = SD(filename)
            for var_name, var_info in sd.datasets().items():
                if var_info[1] == valid_shape:
                    variables.add(var_name)

        return variables

    def _create_coord_list(self, filenames, index_offset=0):
        from cis.data_io.hdf_vd import get_data
        from cis.data_io.hdf_vd import VDS
        from pyhdf.error import HDF4Error
        from cis.data_io import hdf_sd
        import datetime as dt
        from cis.time_util import convert_sec_since_to_std_time, cis_standard_time_unit

        variables = ['Latitude', 'Longitude', "Profile_Time", "Pressure"]
        logging.info("Listing coordinates: " + str(variables))

        # reading data from files
        sdata = {}
        for filename in filenames:
            try:
                sds_dict = hdf_sd.read(filename, variables)
            except HDF4Error as e:
                raise IOError(str(e))

            for var in list(sds_dict.keys()):
                utils.add_element_to_list_in_dict(sdata, var, sds_dict[var])

        alt_name = "altitude"
        logging.info("Additional coordinates: '" + alt_name + "'")

        # work out size of data arrays
        # the coordinate variables will be reshaped to match that.
        # NOTE: This assumes that all Caliop_L1 files have the same altitudes.
        #       If this is not the case, then the following line will need to be changed
        #       to concatenate the data from all the files and not just arbitrarily pick
        #       the altitudes from the first file.
        alt_data = get_data(VDS(filenames[0], "Lidar_Data_Altitudes"), True)
        alt_data *= 1000.0  # Convert to m
        len_x = alt_data.shape[0]

        lat_data = hdf.read_data(sdata['Latitude'], self._get_calipso_data)
        len_y = lat_data.shape[0]

        new_shape = (len_x, len_y)

        # altitude
        alt_data = utils.expand_1d_to_2d_array(alt_data, len_y, axis=0)
        alt_metadata = Metadata(name=alt_name, standard_name=alt_name, shape=new_shape)
        alt_coord = Coord(alt_data, alt_metadata)

        # pressure
        if self.include_pressure:
            pres_data = hdf.read_data(sdata['Pressure'], self._get_calipso_data)
            pres_metadata = hdf.read_metadata(sdata['Pressure'], "SD")
            # Fix badly formatted units which aren't CF compliant and will break if they are aggregated
            if str(pres_metadata.units) == "hPA":
                pres_metadata.units = "hPa"
            pres_metadata.shape = new_shape
            pres_coord = Coord(pres_data, pres_metadata, 'P')

        # latitude
        lat_data = utils.expand_1d_to_2d_array(lat_data[:, index_offset], len_x, axis=1)
        lat_metadata = hdf.read_metadata(sdata['Latitude'], "SD")
        lat_metadata.shape = new_shape
        lat_coord = Coord(lat_data, lat_metadata, 'Y')

        # longitude
        lon = sdata['Longitude']
        lon_data = hdf.read_data(lon, self._get_calipso_data)
        lon_data = utils.expand_1d_to_2d_array(lon_data[:, index_offset], len_x, axis=1)
        lon_metadata = hdf.read_metadata(lon, "SD")
        lon_metadata.shape = new_shape
        lon_coord = Coord(lon_data, lon_metadata, 'X')

        # profile time, x
        time = sdata['Profile_Time']
        time_data = hdf.read_data(time, self._get_calipso_data)
        time_data = convert_sec_since_to_std_time(time_data, dt.datetime(1993, 1, 1, 0, 0, 0))
        time_data = utils.expand_1d_to_2d_array(time_data[:, index_offset], len_x, axis=1)
        time_coord = Coord(time_data, Metadata(name='Profile_Time', standard_name='time', shape=time_data.shape,
                                               units=cis_standard_time_unit), "T")

        # create the object containing all coordinates
        coords = CoordList()
        coords.append(lat_coord)
        coords.append(lon_coord)
        coords.append(time_coord)
        coords.append(alt_coord)
        if self.include_pressure and (pres_data.shape == alt_data.shape):
            # For MODIS L1 this may is not be true, so skips the air pressure reading. If required for MODIS L1 then
            # some kind of interpolation of the air pressure would be required, as it is on a different (smaller) grid
            # than for the Lidar_Data_Altitudes.
            coords.append(pres_coord)

        return coords

    def create_coords(self, filenames, variable=None):
        return UngriddedCoordinates(self._create_coord_list(filenames))

    def create_data_object(self, filenames, variable):
        '''
        To be implemented by subclass
        :param filenames:
        :param variable:
        :return:
        '''
        return None

    def _get_calipso_data(self, sds):
        """
        Reads raw data from an SD instance. Automatically applies the
        scaling factors and offsets to the data arrays found in Calipso data.

        Returns:
            A numpy array containing the raw data with missing data is replaced by NaN.

        Arguments:
            sds        -- The specific sds instance to read

        """
        from cis.utils import create_masked_array_for_missing_data
        import numpy as np

        calipso_fill_values = {'Float_32': -9999.0,
                               # 'Int_8' : 'See SDS description',
                               'Int_16': -9999,
                               'Int_32': -9999,
                               'UInt_8': -127,
                               # 'UInt_16' : 'See SDS description',
                               # 'UInt_32' : 'See SDS description',
                               'ExtinctionQC Fill Value': 32768,
                               'FeatureFinderQC No Features Found': 32767,
                               'FeatureFinderQC Fill Value': 65535}

        data = sds.get()
        attributes = sds.attributes()

        # Missing data. First try 'fillvalue'
        missing_val = attributes.get('fillvalue', None)
        if missing_val is None:
            try:
                # Now try and lookup the fill value based on the data type
                missing_val = calipso_fill_values[attributes.get('format', None)]
            except KeyError:
                # Last guess
                missing_val = attributes.get('_FillValue', None)

        if missing_val is not None:
            data = create_masked_array_for_missing_data(data, missing_val)

        # Now handle valid range mask
        valid_range = attributes.get('valid_range', None)
        if valid_range is not None:
            # Split the range into two numbers of the right type
            v_range = np.asarray(valid_range.split("..."), dtype=data.dtype)
            # Some valid_ranges appear to have only one value, so ignore those...
            if len(v_range) == 2:
                logging.debug("Masking all values {} > v > {}.".format(*v_range))
                data = np.ma.masked_outside(data, *v_range)
            else:
                logging.warning("Invalid valid_range: {}. Not masking values.".format(valid_range))

        # Offsets and scaling.
        offset = attributes.get('add_offset', 0)
        scale_factor = attributes.get('scale_factor', 1)
        data = self._apply_scaling_factor_CALIPSO(data, scale_factor, offset)

        return data

    def _apply_scaling_factor_CALIPSO(self, data, scale_factor, offset):
        """
        Apply scaling factor Calipso data.

        This isn't explicitly documented, but is referred to in the CALIOP docs here:
        http://www-calipso.larc.nasa.gov/resources/calipso_users_guide/data_summaries/profile_data.php#cloud_layer_fraction

        And also confirmed by email with jason.l.tackett@nasa.gov

        :param data:
        :param scale_factor:
        :param offset:
        :return:
        """
        logging.debug("Applying 'science_data = (packed_data / {scale}) + {offset}' "
                      "transformation to data.".format(scale=scale_factor, offset=offset))
        return (data / scale_factor) + offset


class Caliop_L2(abstract_Caliop):
    def get_file_signature(self):
        return [r'CAL_LID_L2_05kmAPro.*hdf']

    def _get_mixed_resolution_calipso_data(self, sds):
        # The first slice of the last dimension corresponds to the higher (in altitude) element of each
        # sub-Level-2-resolution bin.
        return self._get_calipso_data(sds)[:, :, 0]

    def create_coords(self, filenames, variable=None):
        return UngriddedCoordinates(super(Caliop_L2, self)._create_coord_list(filenames, index_offset=1))

    def _create_one_dimensional_coord_list(self, filenames, index_offset=1):
        """
        Create a set of coordinates appropriate for a ond-dimensional (column integrated) variable
        :param filenames:
        :param int index_offset: For 5km products this will choose the coordinates which represent the start (0),
        middle (1) and end (2) of the 15 shots making up each column retrieval.
        :return:
        """
        from pyhdf.error import HDF4Error
        from cis.data_io import hdf_sd
        import datetime as dt
        from cis.time_util import convert_sec_since_to_std_time, cis_standard_time_unit

        variables = ['Latitude', 'Longitude', "Profile_Time"]
        logging.info("Listing coordinates: " + str(variables))

        # reading data from files
        sdata = {}
        for filename in filenames:
            try:
                sds_dict = hdf_sd.read(filename, variables)
            except HDF4Error as e:
                raise IOError(str(e))

            for var in list(sds_dict.keys()):
                utils.add_element_to_list_in_dict(sdata, var, sds_dict[var])

        # latitude
        lat_data = hdf.read_data(sdata['Latitude'], self._get_calipso_data)[:, index_offset]
        lat_metadata = hdf.read_metadata(sdata['Latitude'], "SD")
        lat_coord = Coord(lat_data, lat_metadata, 'Y')

        # longitude
        lon = sdata['Longitude']
        lon_data = hdf.read_data(lon, self._get_calipso_data)[:, index_offset]
        lon_metadata = hdf.read_metadata(lon, "SD")
        lon_coord = Coord(lon_data, lon_metadata, 'X')

        # profile time, x
        time = sdata['Profile_Time']
        time_data = hdf.read_data(time, self._get_calipso_data)[:, index_offset]
        time_data = convert_sec_since_to_std_time(time_data, dt.datetime(1993, 1, 1, 0, 0, 0))
        time_coord = Coord(time_data, Metadata(name='Profile_Time', standard_name='time', shape=time_data.shape,
                                               units=cis_standard_time_unit), "T")

        # create the object containing all coordinates
        coords = CoordList()
        coords.append(lat_coord)
        coords.append(lon_coord)
        coords.append(time_coord)

        return coords

    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        if variable.startswith('Column'):
            coords = self._create_one_dimensional_coord_list(filenames, index_offset=1)
        else:
            coords = self._create_coord_list(filenames, index_offset=1)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        if variable in MIXED_RESOLUTION_VARIABLES:
            logging.warning("Using Level 2 resolution profile for mixed resolution variable {}. See CALIPSO "
                            "documentation for more details".format(variable))
            callback = self._get_mixed_resolution_calipso_data
        else:
            callback = self._get_calipso_data

        return UngriddedData(var, metadata, coords, callback)

    def get_file_format(self, filename):
        return "HDF4/CaliopL2"


class Caliop_L2_NO_PRESSURE(Caliop_L2):
    """
    Read CALIOP L2 data without the pressure coordinate (which is often masked and causes the data to be flattened)
    """
    include_pressure = False


class Caliop_L1(abstract_Caliop):
    def get_file_signature(self):
        return [r'CAL_LID_L1.*hdf']

    def create_data_object(self, filenames, variable):
        logging.debug("Creating data object for variable " + variable)

        # reading coordinates
        # the variable here is needed to work out whether to apply interpolation to the lat/lon data or not
        coords = self._create_coord_list(filenames)

        # reading of variables
        sdata, vdata = hdf.read(filenames, variable)

        # retrieve data + its metadata
        var = sdata[variable]
        metadata = hdf.read_metadata(var, "SD")

        return UngriddedData(var, metadata, coords, self._get_calipso_data)

    def get_file_format(self, filename):
        return "HDF4/CaliopL1"



"""
A basic line plot
"""
from cis.plotting.genericplot import GenericPlot


class LinePlot(GenericPlot):
    line_styles = ["solid", "dashed", "dashdot", "dotted"]

    def __call__(self, ax):
        """
        Plots one line
        """
        from cis.exceptions import InvalidDimensionError
        # Translate our argument names to mpl line kwargs
        self.mplkwargs["linewidth"] = self.itemwidth

        if self.itemstyle:
            if self.itemstyle in LinePlot.line_styles:
                self.mplkwargs["linestyle"] = self.itemstyle
            else:
                from cis.exceptions import InvalidLineStyleError
                raise InvalidLineStyleError(
                    "'" + self.itemstyle + "' is not a valid line style, please use one of: " + str(
                        self.line_styles))

        if self.color:
            self.mplkwargs["color"] = self.color

        if self.x.shape[0] != self.data.shape[0]:
            raise InvalidDimensionError("The plot axes are incompatible, please check and specify at least one "
                                        "axis manually.")

        ax.plot(self.x, self.data, *self.mplargs, **self.mplkwargs)

        super(LinePlot, self).__call__(ax)



"""
A scatter plot with one dataset plotted against another (as opposed to plotting data against a coordinate)
"""
import logging
from cis.plotting.genericplot import APlot


class ComparativeScatter(APlot):

    def __init__(self, packed_data, *mplargs, **mplkwargs):
        """
        Note that the packed_data argument is ignored for this plot type - only xaxis and yaxis are plotted.
        """
        super(ComparativeScatter, self).__init__(packed_data, *mplargs, **mplkwargs)

        logging.debug("Unpacking the data items")
        self.x, self.y = getattr(self.xaxis, 'points', None) or self.xaxis.data, packed_data.data

        self.label = packed_data.long_name if self.label is None else self.label

        self.xlabel = self.xaxis.name()
        self.ylabel = packed_data.name()

    def __call__(self, ax):
        if self.itemwidth is not None:
            self.mplkwargs["markersize"] = self.itemwidth

        if self.itemstyle is not None:
            self.mplkwargs["marker"] = self.itemstyle

        if self.edgecolor is not None:
            self.mplkwargs["edgecolors"] = self.edgecolor

        if self.color is not None:
            self.mplkwargs["c"] = self.color

        if self.label is not None:
            self.mplkwargs['label'] = self.label

        ax.plot(self.x.flatten(), self.y.flatten(), 'o', *self.mplargs, **self.mplkwargs)

        self._plot_xy_line(ax)

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

    def _plot_xy_line(self, ax):
        """
        Plot an x-y dashed line as a guide for the eye
        """
        import numpy as np
        from cis.utils import no_autoscale

        # Turn the scaling off so that we don't change the limits by plotting the line
        with no_autoscale(ax):
            lims = [np.min([ax.get_xlim(), ax.get_ylim()]),  # min of both axes
                    np.max([ax.get_xlim(), ax.get_ylim()])]  # max of both axes

            # now plot both limits against each other for the x=y line
            ax.plot(lims, lims, color="black", linestyle="dashed")

    def is_map(self):
        return False



# The TestRegularGridInterpolator class was adapted from the SciPy original here:
# https://github.com/scipy/scipy/blob/master/scipy/interpolate/tests/test_interpolate.py

# Copyright (c) 2001, 2002 Enthought, Inc.
# All rights reserved.
#
# Copyright (c) 2003-2013 SciPy Developers.
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
#  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# We've changed the order in which the points and sample get passed as per our API change, but otherwise the tests are
# the same

from __future__ import division, print_function, absolute_import

import itertools
import numpy as np
from numpy.testing import (assert_array_almost_equal, assert_raises,
                           TestCase, assert_allclose)

from cis.collocation.gridded_interpolation import _RegularGridInterpolator, GriddedUngriddedInterpolator
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator


class TestRegularGridInterpolator(TestCase):

    def _get_sample_4d(self):
        # create a 4d grid of 3 points in each dimension
        points = [(0., .5, 1.)] * 4
        values = np.asarray([0., .5, 1.])
        values0 = values[:, np.newaxis, np.newaxis, np.newaxis]
        values1 = values[np.newaxis, :, np.newaxis, np.newaxis]
        values2 = values[np.newaxis, np.newaxis, :, np.newaxis]
        values3 = values[np.newaxis, np.newaxis, np.newaxis, :]
        values = (values0 + values1 * 10 + values2 * 100 + values3 * 1000)
        return points, values

    def _get_sample_4d_2(self):
        # create another 4d grid of 3 points in each dimension
        points = [(0., .5, 1.)] * 2 + [(0., 5., 10.)] * 2
        values = np.asarray([0., .5, 1.])
        values0 = values[:, np.newaxis, np.newaxis, np.newaxis]
        values1 = values[np.newaxis, :, np.newaxis, np.newaxis]
        values2 = values[np.newaxis, np.newaxis, :, np.newaxis]
        values3 = values[np.newaxis, np.newaxis, np.newaxis, :]
        values = (values0 + values1 * 10 + values2 * 100 + values3 * 1000)
        return points, values

    def test_list_input(self):
        points, values = self._get_sample_4d()

        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8],
                                          [0.5, 0.5, .5, .5]])

        for method in ['lin', 'nn']:
            interp = _RegularGridInterpolator(points,
                                             sample.T.tolist(), method=method)
            v1 = interp(values.tolist())
            interp = _RegularGridInterpolator(points,
                                             sample.T, method=method)
            v2 = interp(values)
            assert_allclose(v1, v2)

    def test_complex(self):
        points, values = self._get_sample_4d()
        values = values - 2j*values
        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8], [0.5, 0.5, .5, .5]])

        for method in ['lin', 'nn']:
            interp = _RegularGridInterpolator(points, sample.T, method=method)
            rinterp = _RegularGridInterpolator(points, sample.real.T, method=method)
            iinterp = _RegularGridInterpolator(points, sample.imag.T, method=method)

            v1 = interp(values)
            v2 = rinterp(values) + 1j*iinterp(values)
            assert_allclose(v1, v2)

    def test_linear_xi1d(self):
        points, values = self._get_sample_4d_2()
        sample = np.asarray([0.1, 0.1, 10., 9.])
        wanted = 1001.1
        interp = _RegularGridInterpolator(points, sample.T)
        assert_array_almost_equal(interp(values), wanted)

    def test_linear_xi3d(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8],
                                                     [0.5, 0.5, .5, .5]])
        wanted = np.asarray([1001.1, 846.2, 555.5])
        interp = _RegularGridInterpolator(points, sample.T)
        assert_array_almost_equal(interp(values), wanted)

    def test_nearest(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([0.1, 0.1, .9, .9])
        wanted = 1100.
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        assert_array_almost_equal(interp(values), wanted)
        sample = np.asarray([0.1, 0.1, 0.1, 0.1])
        wanted = 0.
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        assert_array_almost_equal(interp(values), wanted)
        sample = np.asarray([0., 0., 0., 0.])
        wanted = 0.
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        assert_array_almost_equal(interp(values), wanted)
        sample = np.asarray([1., 1., 1., 1.])
        wanted = 1111.
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        assert_array_almost_equal(interp(values), wanted)
        sample = np.asarray([0.1, 0.4, 0.6, 0.9])
        wanted = 1055.
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        assert_array_almost_equal(interp(values), wanted)

    def test_linear_edges(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[0., 0., 0., 0.], [1., 1., 1., 1.]])
        wanted = np.asarray([0., 1111.])
        interp = _RegularGridInterpolator(points, sample.T)
        assert_array_almost_equal(interp(values), wanted)

    def test_valid_create(self):
        # create a 2d grid of 3 points in each dimension
        points = [(0., .5, 1.), (0., 1., .5)]
        values = np.asarray([0., .5, 1.])
        values0 = values[:, np.newaxis]
        values1 = values[np.newaxis, :]
        values = (values0 + values1 * 10)
        assert_raises(ValueError, _RegularGridInterpolator, points, values)
        points = [((0., .5, 1.), ), (0., .5, 1.)]
        assert_raises(ValueError, _RegularGridInterpolator, points, values)
        points = [(0., .5, .75, 1.), (0., .5, 1.)]
        assert_raises(ValueError, _RegularGridInterpolator, points, values)

    def test_valid_call(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[0., 0., 0., 0.], [1., 1., 1., 1.]])
        interp = _RegularGridInterpolator(points, sample.T)
        assert_raises(ValueError, interp, values, "undefmethod")
        # This is actually an error on creation now
        sample = np.asarray([[0., 0., 0.], [1., 1., 1.]])
        assert_raises(ValueError, _RegularGridInterpolator, points, sample)

    def test_out_of_bounds_extrap(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[-.1, -.1, -.1, -.1], [1.1, 1.1, 1.1, 1.1],
                             [21, 2.1, -1.1, -11], [2.1, 2.1, -1.1, -1.1]])
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        wanted = np.asarray([0., 1111., 11., 11.])
        assert_array_almost_equal(interp(values, fill_value=None), wanted)
        interp = _RegularGridInterpolator(points, sample.T, method="lin")
        wanted = np.asarray([-111.1, 1222.1, -11068., -1186.9])
        assert_array_almost_equal(interp(values, fill_value=None), wanted)

    def test_out_of_bounds_extrap2(self):
        points, values = self._get_sample_4d_2()
        sample = np.asarray([[-.1, -.1, -.1, -.1], [1.1, 1.1, 1.1, 1.1],
                             [21, 2.1, -1.1, -11], [2.1, 2.1, -1.1, -1.1]])

        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        wanted = np.asarray([0., 11., 11., 11.])
        assert_array_almost_equal(interp(values, fill_value=None), wanted)
        interp = _RegularGridInterpolator(points, sample.T, method="lin")
        wanted = np.asarray([-12.1, 133.1, -1069., -97.9])
        assert_array_almost_equal(interp(values, fill_value=None), wanted)

    def test_out_of_bounds_fill(self):
        points, values = self._get_sample_4d()

        sample = np.asarray([[-.1, -.1, -.1, -.1], [1.1, 1.1, 1.1, 1.1],
                             [2.1, 2.1, -1.1, -1.1]])
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        # Assert that all of the elements are masked
        assert all(interp(values, fill_value=np.nan).mask)
        interp = _RegularGridInterpolator(points, sample.T, method="lin")
        assert all(interp(values, fill_value=np.nan).mask)

        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8],
                             [0.5, 0.5, .5, .5]])
        wanted = np.asarray([1001.1, 846.2, 555.5])
        interp = _RegularGridInterpolator(points, sample.T)
        assert_array_almost_equal(interp(values, fill_value=np.nan), wanted)

    def test_nearest_compare_qhull(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8],
                             [0.5, 0.5, .5, .5]])
        interp = _RegularGridInterpolator(points, sample.T, method="nn")
        points_qhull = itertools.product(*points)
        points_qhull = [p for p in points_qhull]
        points_qhull = np.asarray(points_qhull)
        values_qhull = values.reshape(-1)
        interp_qhull = NearestNDInterpolator(points_qhull, values_qhull)
        assert_array_almost_equal(interp(values), interp_qhull(sample))

    def test_linear_compare_qhull(self):
        points, values = self._get_sample_4d()
        sample = np.asarray([[0.1, 0.1, 1., .9], [0.2, 0.1, .45, .8],
                             [0.5, 0.5, .5, .5]])
        interp = _RegularGridInterpolator(points, sample.T)
        points_qhull = itertools.product(*points)
        points_qhull = [p for p in points_qhull]
        points_qhull = np.asarray(points_qhull)
        values_qhull = values.reshape(-1)
        interp_qhull = LinearNDInterpolator(points_qhull, values_qhull)
        assert_array_almost_equal(interp(values), interp_qhull(sample))

    def test_2d_cube(self):
        from cis.test.util.mock import make_square_5x3_2d_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint

        cube = make_square_5x3_2d_cube()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(1.0, 1.0), HyperPoint(4.0, 4.0), HyperPoint(-4.0, -4.0)])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        values = interpolator(cube)
        wanted = np.asarray([8.8, 11.2, 4.8])
        assert_array_almost_equal(values, wanted)

    def test_hybrid_Coord_nn(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint
        import datetime as dt
        cube = make_mock_cube(time_dim_length=3, hybrid_pr_len=10)

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, pres=111100040.5, alt=5000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=0.0, lon=0.0, pres=113625040.5, alt=4000, t=dt.datetime(1984, 8, 28, 12, 0, 0)),
             HyperPoint(lat=5.0, lon=2.5, pres=177125044.5, alt=3000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=-4.0, lon=-4.0, pres=166600039.0, alt=3500, t=dt.datetime(1984, 8, 27))])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'nn')
        values = interpolator(cube)
        wanted = np.ma.masked_invalid(np.ma.array([221.0, 223.0, 318.0, np.nan]))
        assert_array_almost_equal(values, wanted)

    def test_hybrid_Coord_without_hybrid_coord(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint
        import datetime as dt
        cube = make_mock_cube(time_dim_length=3)

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1984, 8, 28, 12, 0, 0)),
             HyperPoint(lat=5.0, lon=2.5, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=-4.0, lon=-4.0, t=dt.datetime(1984, 8, 27))])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        values = interpolator(cube)
        wanted = np.asarray([23.0, 23.5, 33.5, 12.4])
        assert_array_almost_equal(values, wanted)

    def test_hybrid_Coord(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint
        import datetime as dt
        cube = make_mock_cube(time_dim_length=3, hybrid_pr_len=10)

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, pres=111100040.5, alt=5000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=0.0, lon=0.0, pres=113625040.5, alt=4000, t=dt.datetime(1984, 8, 28, 12, 0, 0)),
             HyperPoint(lat=5.0, lon=2.5, pres=177125044.5, alt=3000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=-4.0, lon=-4.0, pres=166600039.0, alt=3500, t=dt.datetime(1984, 8, 27))])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        values = interpolator(cube)
        wanted = np.ma.masked_invalid(np.ma.array([221.5, 226.5, 330.5, np.nan]))
        assert_array_almost_equal(values, wanted)

    def test_hybrid_pres_Coord_order_doesnt_matter(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint
        import datetime as dt
        cube = make_mock_cube(time_dim_length=3, hybrid_pr_len=10)

        cube.transpose()

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, pres=111100040.5, alt=5000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=0.0, lon=0.0, pres=113625040.5, alt=4000, t=dt.datetime(1984, 8, 28, 12, 0, 0)),
             HyperPoint(lat=5.0, lon=2.5, pres=177125044.5, alt=3000, t=dt.datetime(1984, 8, 28, 0, 0, 0)),
             HyperPoint(lat=-4.0, lon=-4.0, pres=166600039.0, alt=3500, t=dt.datetime(1984, 8, 27))])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        values = interpolator(cube)
        wanted = np.ma.masked_invalid(np.ma.array([221.5, 226.5, 330.5, np.nan]))
        assert_array_almost_equal(values, wanted)

    def test_hybrid_alt_Coord_order_doesnt_matter(self):
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint
        import datetime as dt
        cube = make_mock_cube(time_dim_length=3, hybrid_ht_len=10)

        cube.transpose()

        sample_points = UngriddedData.from_points_array(
            # This point actually lies outside the lower bounds for altitude at this point in space
            [HyperPoint(lat=1.0, lon=1.0, alt=5000.0, t=dt.datetime(1984, 8, 28, 8, 34)),
             # This point lies in the middle of the altitude bounds at this point
             HyperPoint(lat=4.0, lon=4.0, alt=6000.0, t=dt.datetime(1984, 8, 28, 8, 34)),
             # This point lies outside the upper bounds for altitude at this point
             HyperPoint(lat=-4.0, lon=-4.0, alt=6500.0, t=dt.datetime(1984, 8, 27, 2, 18, 52))])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'nn')
        values = interpolator(cube, fill_value=None)
        wanted = np.asarray([221.0, 345.0, 100.0])
        assert_array_almost_equal(values, wanted)

    def test_invalid_fill_value(self):
        np.random.seed(1234)
        x = np.linspace(0, 2, 5)
        y = np.linspace(0, 1, 7)
        values = np.random.rand(5, 7)

        # integers can be cast to floats
        interp = _RegularGridInterpolator((x, y), [0.4, 0.7])
        interp(values, fill_value=1)

        # complex values cannot
        assert_raises(ValueError, interp, values, fill_value=1+2j)

    def test_fillvalue_type(self):
        # from #3703; test that interpolator object construction succeeds
        values = np.ones((10, 20, 30), dtype='>f4')
        points = [np.arange(n) for n in values.shape]
        xi = np.asarray([1, 1, 1])
        interpolator = _RegularGridInterpolator(points, xi.T)
        interpolator(values, fill_value=0.)

    def test_missing_values(self):
        from cis.test.util.mock import make_square_5x3_2d_cube_with_missing_data

        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint

        cube = make_square_5x3_2d_cube_with_missing_data()

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0),
             HyperPoint(lat=0.0, lon=4.0),
             HyperPoint(lat=-4.0, lon=1.0)])

        wanted = np.ma.array([8.0, np.nan, np.nan], mask=[False, True, True])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'nn')
        assert_array_almost_equal(interpolator(cube), wanted)

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        assert_array_almost_equal(interpolator(cube), wanted)

    def test_missing_values_with_extrapolation(self):
        from cis.test.util.mock import make_square_5x3_2d_cube_with_missing_data

        from cis.data_io.ungridded_data import UngriddedData
        from cis.data_io.hyperpoint import HyperPoint

        cube = make_square_5x3_2d_cube_with_missing_data()

        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=-5.0, lon=10.0),
             HyperPoint(lat=0.0, lon=10.0),
             HyperPoint(lat=-4.0, lon=1.0)])

        wanted = np.ma.array([6.0, np.nan, np.nan], mask=[False, True, True])

        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'nn')
        assert_array_almost_equal(interpolator(cube, extrapolate=True), wanted)

        wanted = np.ma.array([7.0, np.nan, np.nan], mask=[False, True, True])
        interpolator = GriddedUngriddedInterpolator(cube, sample_points, 'lin')
        assert_array_almost_equal(interpolator(cube, extrapolate=True), wanted)


class MyValue(object):
    """
    Minimal indexable object
    """

    def __init__(self, shape):
        self.ndim = 2
        self.shape = shape
        self._v = np.arange(np.prod(shape)).reshape(shape)

    def __getitem__(self, idx):
        return self._v[idx]

    def __array_interface__(self):
        return None

    def __array__(self):
        raise RuntimeError("No array representation")



"""
module to test the time conversion utilities
"""
from nose.tools import eq_, assert_almost_equal, istest, raises
import datetime as dt
from unittest import TestCase

from cis.time_util import convert_sec_since_to_std_time, convert_datetime_to_std_time, \
    convert_julian_date_to_std_time, PartialDateTime, find_last_day_of_month, set_year, change_year_of_ungridded_data


class TestTimeUtils(TestCase):

    def test_that_can_convert_tai_to_datetime_obj(self):
        import numpy as np
        sec = 1.0/(24.0*60.0*60.0)
        days_since_standard_epoch = 143541.0  # Almost, but not quite 365.2425*393.0, not sure why...

        a = np.arange(6).reshape(2, 3)
        b = convert_sec_since_to_std_time(a, dt.datetime(1993, 1, 1))

        eq_(a.shape, b.shape)
        assert_almost_equal(b[0][0], days_since_standard_epoch)
        assert_almost_equal(b[0][1], days_since_standard_epoch+1*sec)
        assert_almost_equal(b[0][2], days_since_standard_epoch+2*sec)
        assert_almost_equal(b[1][0], days_since_standard_epoch+3*sec)
        assert_almost_equal(b[1][1], days_since_standard_epoch+4*sec)
        assert_almost_equal(b[1][2], days_since_standard_epoch+5*sec)

    def test_that_can_convert_masked_tai_to_datetime_obj(self):
        import numpy.ma as ma
        sec = 1.0/(24.0*60.0*60.0)
        days_since_standard_epoch = 143541.0  # Almost, but not quite 365.2425*393.0, not sure why...

        a = ma.array([0, 1, 2, 3, 4, 5], mask=[False, False, True, False, False, False]).reshape(2, 3)
        b = convert_sec_since_to_std_time(a, dt.datetime(1993, 1, 1))

        eq_(a.shape, b.shape)
        assert_almost_equal(b[0][0], days_since_standard_epoch)
        assert_almost_equal(b[0][1], days_since_standard_epoch+1*sec)
        assert_almost_equal(b.filled()[0][2], b.fill_value)
        assert_almost_equal(b[1][0], days_since_standard_epoch+3*sec)
        assert_almost_equal(b[1][1], days_since_standard_epoch+4*sec)
        assert_almost_equal(b[1][2], days_since_standard_epoch+5*sec)

    def test_that_can_convert_julian_tai_to_datetime_obj(self):
        import numpy as np
        sec = 1.0/(24.0*60.0*60.0)
        days_since_standard_epoch = 143541.0  # Almost, but not quite 365.2425*393.0, not sure why...

        a = np.arange(6).reshape(2, 3)
        b = convert_sec_since_to_std_time(a, dt.datetime(1993, 1, 1))

        eq_(a.shape, b.shape)
        assert_almost_equal(b[0][0], days_since_standard_epoch)
        assert_almost_equal(b[0][1], days_since_standard_epoch+1*sec)
        assert_almost_equal(b[0][2], days_since_standard_epoch+2*sec)
        assert_almost_equal(b[1][0], days_since_standard_epoch+3*sec)
        assert_almost_equal(b[1][1], days_since_standard_epoch+4*sec)
        assert_almost_equal(b[1][2], days_since_standard_epoch+5*sec)

    def test_that_can_calculate_mid_point_between_two_datetime(self):
        from cis.time_util import calculate_mid_time
        t1 = convert_datetime_to_std_time(dt.datetime(2010, 0o2, 0o5, 0, 0, 0))
        t2 = convert_datetime_to_std_time(dt.datetime(2010, 0o2, 0o6, 0, 0, 0))
        tm = calculate_mid_time(t1, t2)
        eq_(tm, convert_datetime_to_std_time(dt.datetime(2010, 0o2, 0o5, 12, 0, 0)))

    def test_convert_julian_date_to_std_time(self):
        import numpy as np
        from cis.time_util import convert_datetime_to_std_time

        julian_days = np.array([2454637.8091, 2454637.8092, 2454637.8097,
                                2454637.8197, 2454638.8097, 2454657.8097,
                                2454737.8197, 2456638.8097, 2464657.8097])

        std_days = convert_julian_date_to_std_time(julian_days)

        ref = convert_datetime_to_std_time([dt.datetime(2008, 6, 20, 7, 25, 6),
                                            dt.datetime(2008, 6, 20, 7, 25, 15),
                                            dt.datetime(2008, 6, 20, 7, 25, 58),
                                            dt.datetime(2008, 6, 20, 7, 40, 22),
                                            dt.datetime(2008, 6, 21, 7, 25, 58),
                                            dt.datetime(2008, 7, 10, 7, 25, 58),
                                            dt.datetime(2008, 9, 28, 7, 40, 22),
                                            dt.datetime(2013, 12, 12, 7, 25, 58),
                                            dt.datetime(2035, 11, 26, 7, 25, 58)])

        eq_(julian_days.shape, std_days.shape)
        assert np.allclose(std_days, ref)

    # Tests for find_last_day_of_month
    def test_find_last_day_of_month_finds_day_for_dec_2010(self):
        day = find_last_day_of_month(2010, 12)
        assert (day == 31)

    def test_find_last_day_of_month_finds_day_for_feb_2000(self):
        day = find_last_day_of_month(2000, 2)
        assert (day == 29)

    # Tests for convert_datetime_components_to_datetime
    def test_convert_datetime_components_to_datetime_can_convert_year_as_lower_limit(self):
        start, end = PartialDateTime(2000).range()
        assert (start == dt.datetime(2000, 1, 1, 0, 0, 0))
        assert (end == dt.datetime(2000, 12, 31, 23, 59, 59))

    def test_convert_datetime_components_to_datetime_can_convert_year_month_as_lower_limit(self):
        start, end = PartialDateTime(1990, 6).range()
        assert (start == dt.datetime(1990, 6, 1, 0, 0, 0))
        assert (end == dt.datetime(1990, 6, 30, 23, 59, 59))

    def test_convert_datetime_components_to_datetime_can_convert_year_month_day_as_lower_limit(self):
        start, end = PartialDateTime(1990, 6, 7).range()
        assert (start == dt.datetime(1990, 6, 7, 0, 0, 0))
        assert (end == dt.datetime(1990, 6, 7, 23, 59, 59))

    def test_convert_datetime_components_to_datetime_can_convert_date_hour_as_lower_limit(self):
        start, end = PartialDateTime(1990, 6, 7, 18).range()
        assert (start == dt.datetime(1990, 6, 7, 18, 0, 0))
        assert (end == dt.datetime(1990, 6, 7, 18, 59, 59))

    def test_convert_datetime_components_to_datetime_can_convert_date_hour_min_as_lower_limit(self):
        start, end = PartialDateTime(1990, 6, 7, 6, 30).range()
        assert (start == dt.datetime(1990, 6, 7, 6, 30, 0))
        assert (end == dt.datetime(1990, 6, 7, 6, 30, 59))

    def test_convert_datetime_components_to_datetime_can_convert_date_hour_min_sec_as_lower_limit(self):
        start, end = PartialDateTime(1990, 6, 7, 12, 15, 45).range()
        assert (start == dt.datetime(1990, 6, 7, 12, 15, 45))
        assert (end == dt.datetime(1990, 6, 7, 12, 15, 45))

    @raises(ValueError)
    def test_convert_datetime_components_to_datetime_raises_error_if_invalid_date(self):
        start, end = PartialDateTime(2000, 6, 31).range()

    @raises(ValueError)
    def test_convert_datetime_components_to_datetime_raises_error_if_invalid_time(self):
        start, end = PartialDateTime(2000, 6, 30, 12, 30, 60).range()

    def test_set_year(self):
        from datetime import datetime

        # Test changing a leapday to a year without a leapday returns None
        eq_(set_year(datetime(1984, 2, 28), 2007), datetime(2007, 2, 28))
        eq_(set_year(datetime(1984, 2, 29), 2007), None)

        # Test changing a leapday to a year with a leapday works fine
        eq_(set_year(datetime(1984, 2, 28), 2000), datetime(2000, 2, 28))
        eq_(set_year(datetime(1984, 2, 29), 2000), datetime(2000, 2, 29))

    def test_change_ug_data_year(self):
        from cis.test.util.mock import make_regular_2d_with_time_ungridded_data
        from cis.time_util import convert_datetime_to_std_time
        from datetime import datetime

        ug = make_regular_2d_with_time_ungridded_data()

        change_year_of_ungridded_data(ug, 2007)

        eq_(ug.coord('time').points[0, 0], convert_datetime_to_std_time(datetime(2007, 8, 27)))



import unittest
import os
from netCDF4 import Dataset

from cis.data_io.gridded_data import make_from_cube, GriddedDataList
from cis.test.util.mock import make_dummy_2d_ungridded_data, make_mock_cube
from cis.test.integration_test_data import valid_cis_col_file, valid_cis_col_variable
from cis.data_io.write_netcdf import write

tmp_file = "tmp_file.nc"


class TestWriteNetcdf(unittest.TestCase):

    def tearDown(self):
        if hasattr(self, 'd') and self.d.isopen():
            self.d.close()
        os.remove(tmp_file)

    def test_write_netcdf(self):
        data_object = make_dummy_2d_ungridded_data()
        write(data_object, tmp_file)

    def test_write_col_and_reload_1(self):
        # Copy a collocated file and try to reload it.  This exposes a bug where
        # var.shape is set in the NetCDF metadata
        from cis.data_io.products import Aerosol_CCI_L2

        prod = Aerosol_CCI_L2()
        data_object = prod.create_data_object([valid_cis_col_file], valid_cis_col_variable)
        write(data_object, tmp_file)

        self.d = Dataset(tmp_file)

        v = self.d.variables['AOT_440']

        # This will fail because var.shape is in the file
        print(v[:2])

    def test_write_col_and_reload_2(self):
        # Copy a collocated file and try to reload it.  This exposes a bug where
        # latitude and longitude aren't recognised on reload
        from cis.data_io.products import cis

        prod = cis()
        data_object = prod.create_data_object([valid_cis_col_file], valid_cis_col_variable)
        write(data_object, tmp_file)

        data_object2 = prod.create_data_object([tmp_file], valid_cis_col_variable)

    def test_ungridded_write_attributes(self):
        data = make_dummy_2d_ungridded_data()
        attrs = {'attr_name': 'attr_val',
                 'standard_name': 'std_val',
                 'long_name': 'lg_val',
                 'units': 'units'}
        data.add_attributes(attrs)
        write(data, tmp_file)
        self.d = Dataset(tmp_file)
        for key, val in attrs.items():
            assert getattr(self.d.variables['rainfall_flux'], key) == val

    def test_gridded_write_attributes(self):
        data = make_from_cube(make_mock_cube())
        data.var_name = 'rain'
        attrs = {'attr_name': 'attr_val',
                 'standard_name': 'convective_rainfall_amount',
                 'long_name': 'lg_val',
                 'units': 'units'}
        data.add_attributes(attrs)
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        for key, val in attrs.items():
            assert getattr(self.d.variables['rain'], key) == val

    def test_ungridded_write_units(self):
        data = make_dummy_2d_ungridded_data()
        data.units = 'kg'
        write(data, tmp_file)
        self.d = Dataset(tmp_file)
        assert self.d.variables['rainfall_flux'].units == 'kg'

    def test_gridded_write_units(self):
        data = make_from_cube(make_mock_cube())
        data.var_name = 'rain'
        data.units = 'ppm'
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        assert self.d.variables['rain'].units == 'ppm'

    def test_gridded_write_time_as_unlimited_dimension(self):
        data = make_from_cube(make_mock_cube(time_dim_length=7))
        data.var_name = 'rain'
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        assert self.d.dimensions['time'].isunlimited()

    def test_gridded_write_no_time_has_no_unlimited_dimension(self):
        data = make_from_cube(make_mock_cube())
        data.var_name = 'rain'
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        for d in self.d.dimensions.values():
            assert not d.isunlimited()

    def test_gridded_list_write_time_as_unlimited_dimension(self):
        data = GriddedDataList([make_from_cube(make_mock_cube(time_dim_length=7))])
        data[0].var_name = 'rain'
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        assert self.d.dimensions['time'].isunlimited()

    def test_gridded_list_write_no_time_has_no_unlimited_dimension(self):
        data = GriddedDataList([make_from_cube(make_mock_cube())])
        data[0].var_name = 'rain'
        data.save_data(tmp_file)
        self.d = Dataset(tmp_file)
        for d in self.d.dimensions.values():
            assert not d.isunlimited()

    # def test_can_write_hierarchical_group_variables(self):
    #     from cis.test.integration_test_data import valid_nested_groups_file
    #     from cis import read_data
    #     from hamcrest import assert_that, is_
    #     var_name = 'group1/group2/var4'
    #     data = read_data(valid_nested_groups_file, var_name, product='cis')
    #     assert_that(data.data, is_([12321]))
    #     data.save_data(tmp_file)
    #     self.d = Dataset(tmp_file)
    #     assert_that(self.d.variables[var_name][:], is_([12321]))



import logging
import iris
from cis.data_io import gridded_data as gridded_data
from cis.data_io.products import AProduct


class NetCDF_Gridded(AProduct):
    def get_file_signature(self):
        # We don't know of any 'standard' netCDF CF model data yet...
        return [r'.*\.nc']

    @staticmethod
    def load_multiple_files_callback(cube, field, filename):
        # Static callback method for passing to iris when loading a single cube. To be implemented by subclasses as
        # needed
        pass

    @staticmethod
    def load_single_file_callback(cube, field, filename):
        # Static callback method for passing to iris when loading multiple cubes. To be implemented by subclasses as
        # needed
        pass

    def get_variable_names(self, filenames, data_type=None):
        import iris
        import cf_units as unit
        from cis.utils import single_warnings_only

        variables = []

        # Filter the warnings so that they only appear once - otherwise you get lots of repeated warnings
        with single_warnings_only():
            cubes = iris.load(filenames)

        for cube in cubes:
            is_time_lat_lon_pressure_altitude_or_has_only_1_point = True
            for dim in cube.dim_coords:
                units = dim.units
                if dim.points.size > 1 and \
                        not units.is_time() and \
                        not units.is_time_reference() and \
                        not units.is_vertical() and \
                        not units.is_convertible(unit.Unit('degrees')):
                    is_time_lat_lon_pressure_altitude_or_has_only_1_point = False
                    break
            if is_time_lat_lon_pressure_altitude_or_has_only_1_point:
                if cube.var_name:
                    variables.append(cube.var_name)
                else:
                    variables.append(cube.name())

        return set(variables)

    def create_coords(self, filenames, variable=None):
        """Reads the coordinates on which a variable depends.
        Note: This calls create_data_object because the coordinates are returned as a Cube.
        :param filenames: list of names of files from which to read coordinates
        :param variable: name of variable for which the coordinates are required
        (optional if file contains only one cube)
        :return: iris.cube.Cube
        """

        if variable is None:
            variable_names = self.get_variable_names(filenames)
            if len(variable_names) > 1:
                variable_name = str(variable_names.pop())
                logging.debug("Reading an IRIS Cube for the coordinates based on the variable %s" % variable_names)
            else:
                variable_name = None
        else:
            variable_name = variable

        return self.create_data_object(filenames, variable_name)

    def create_data_object(self, filenames, variable):
        """Reads the data for a variable.
        :param filenames: list of names of files from which to read data
        :param variable: (optional) name of variable; if None, the file(s) must contain data for only one cube
        :return: iris.cube.Cube
        """
        from cis.time_util import convert_cube_time_coord_to_standard_time
        from cis.utils import single_warnings_only

        # Filter the warnings so that they only appear once - otherwise you get lots of repeated warnings
        #  - partly because we open the files multiple times (to look for aux coords) and partly because iris
        #  will throw a warning every time it meets a variable with a non-CF dimension
        with single_warnings_only():
            cube = self._create_cube(filenames, variable)

        try:
            cube = convert_cube_time_coord_to_standard_time(cube)
        except iris.exceptions.CoordinateNotFoundError:
            pass
        return cube

    def get_file_format(self, filename):
        return "NetCDF/Gridded"

    def _create_cube(self, filenames, variable):
        """Creates a cube for the specified variable.
        :param filenames: List of filenames to read coordinates from
        :param variable: Optional variable to read while we're reading the coordinates, can be a string or a
        VariableConstraint object
        :return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        from cis.exceptions import InvalidVariableError
        import six

        # Check if the files given actually exist.
        for filename in filenames:
            with open(filename) as f:
                pass

        variable_constraint = variable
        if isinstance(variable, six.string_types):
            variable_constraint = DisplayConstraint(cube_func=(lambda c: c.var_name == variable or
                                                               c.standard_name == variable or
                                                               c.long_name == variable), display=variable)
        if len(filenames) == 1:
            callback_function = self.load_single_file_callback
        else:
            callback_function = self.load_multiple_files_callback

        try:
            cube = gridded_data.load_cube(filenames, variable_constraint, callback=callback_function)
        except ValueError as e:
            if variable is None:
                message = "File contains more than one cube variable name must be specified"
            elif e.args[0] == "No cubes found":
                message = "Variable not found: {} \nTo see a list of variables run: cis info {}" \
                    .format(str(variable), filenames[0])
            else:
                message = e.args[0]
            raise InvalidVariableError(message)

        self._add_available_aux_coords(cube, filenames)

        return cube

    def _add_available_aux_coords(self, cube, filenames):
        """
        Add any altitude or pressure Auxiliary coordinates that IRIS does not load by default.
        :type cube: data_io.gridded_data.GriddedData
        :type filenames: list
        :return:
        """
        from iris.coords import AuxCoord

        def _make_aux_coord_and_dims_from_geopotential(g_cube):
            from scipy.constants import g
            points = g_cube.data / g
            dims = list(range(len(g_cube.shape)))
            coord = AuxCoord(points=points,
                             standard_name='altitude',
                             long_name='Geopotential height at layer midpoints',
                             var_name='altitude',
                             units='meter',
                             attributes={'positive': 'up'})
            return coord, dims

        def _make_aux_coord_and_dims_from_geopotential_height(g_ht_cube):
            dims = list(range(len(g_ht_cube.shape)))
            coord = AuxCoord(points=g_ht_cube.data,
                             standard_name='altitude',
                             long_name='Geopotential height at layer midpoints',
                             var_name='altitude',
                             units='meter',
                             attributes={'positive': 'up'})
            return coord, dims

        def _make_aux_coord_and_dims_from_air_pressure(aps_cube):
            dims = list(range(len(aps_cube.shape)))
            coord = AuxCoord(points=aps_cube.data,
                             standard_name='air_pressure',
                             long_name='Air Pressure',
                             var_name='air_pressure',
                             units='Pa')
            return coord, dims

        aux_coord_creation_functions = {'geopotential': _make_aux_coord_and_dims_from_geopotential,
                                        'geopotential_height': _make_aux_coord_and_dims_from_geopotential_height,
                                        'air_pressure': _make_aux_coord_and_dims_from_air_pressure}

        for standard_name, make_aux_coord in list(aux_coord_creation_functions.items()):
            constraint = DisplayConstraint(cube_func=lambda c: c.standard_name == standard_name,
                                           display=standard_name)
            try:
                aux_cube = gridded_data.load_cube(filenames, constraint)
                aux_coord, dims = make_aux_coord(aux_cube)
                cube.add_aux_coord(aux_coord, dims)
            except ValueError:
                pass  # The field doesn't exist; that's OK we just won't add it.


class DisplayConstraint(iris.Constraint):
    """Variant of iris.Constraint with a string value that can be displayed.
    """

    def __init__(self, *args, **kwargs):
        sc_kwargs = kwargs.copy()
        self.display = str(sc_kwargs.get('display', None))
        if self.display is not None:
            del sc_kwargs['display']
        super(DisplayConstraint, self).__init__(*args, **sc_kwargs)

    def __str__(self):
        if self.display is not None:
            return self.display
        else:
            return super(DisplayConstraint, self).__str__()



"""
A '2D' Histogram plot - e.g. one that has data on both the x and y axis. This is generated using numpy.histogram2d and
a call to pcolormesh. Although strictly a 2D plot it inherits from ComparativeScatter rather than Generic2D because
the unpacking has more in common with that.
"""
from cis.plotting.comparativescatter import ComparativeScatter
import numpy


class Histogram2D(ComparativeScatter):

    def __init__(self, packed_data, logv=None, vstep=None, xbins=10, ybins=10,
                 cbarscale=None, cbarorient=None, colourbar=True, cbarlabel=None, *args, **kwargs):
        """
        Note that the packed_data argument is ignored for this plot type - only xaxis and yaxis are plotted.

        :param CommonData packed_data: IGNORED
        :param bool logv: Log the v (colour bar) axis?
        :param float vstep: The step to use for the v axis (colour bar)
        :param int xbins: The number of histogram bins to use on the xaxis
        :param int ybins: The number of histogram bins to use on the yaxis
        :param bool colourbar: Include colour bar? Default True
        :param float cbarscale: A scale factor for the colorbar
        :param string cbarorient: The colour bar orientation ('vertical' or 'horizontal')
        :param string cbarlabel: A label for the colour bar
        """
        super(Histogram2D, self).__init__(packed_data, *args, **kwargs)

        self.logv = logv
        self.xbins = xbins
        self.ybins = ybins
        self.vstep = vstep
        self.cbarscale = cbarscale
        self.cbarorient = cbarorient
        self.colourbar = colourbar
        self.cbarlabel = cbarlabel or "Frequency"

    def __call__(self, ax):
        """
        Plots a 2d histogram.
        """
        from .plot import add_color_bar
        from cis.utils import apply_intersection_mask_to_two_arrays

        # Numpy histogram2D doesn't like masked arrays, so create the unmased intersection of the two datasets
        first_data_item, second_data_item = apply_intersection_mask_to_two_arrays(self.x, self.y)
        first_data_item = first_data_item.compressed()
        second_data_item = second_data_item.compressed()

        # Use Numpy histogram generator instead of hist2d to allow log scales to be properly plotted
        histogram2ddata, x, y = numpy.histogram2d(first_data_item, second_data_item, bins=[self.xbins, self.ybins])
        histogram2ddata = numpy.ma.masked_equal(histogram2ddata, 0)
        self.map = ax.pcolor(x, y, histogram2ddata.T, *self.mplargs, **self.mplkwargs)

        self._plot_xy_line(ax)

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        if self.colourbar:
            add_color_bar(ax, self.map, self.vstep, self.logv, self.cbarscale, self.cbarorient, self.cbarlabel)



import logging
import six

from cis.data_io.Coord import CoordList
from cis.data_io.products import AProduct, NetCDF_Gridded
from cis.data_io.ungridded_data import UngriddedCoordinates, UngriddedData
from cf_units import Unit, CALENDAR_STANDARD
from abc import ABCMeta, abstractmethod


ESA_TIME_UNIT = Unit("hours since 1970-01-01 00:00:00", calendar=CALENDAR_STANDARD)


@six.add_metaclass(ABCMeta)
class CCI_L2(object):
    """
    Abstract class for the various Level 2 CCI data products. This just defines the interface which
    the subclasses must implement.
    """
    def _create_coord_list(self, filenames):

        from cis.data_io.netcdf import read_many_files_individually, get_metadata
        from cis.data_io.Coord import Coord
        from cis.exceptions import InvalidVariableError

        try:
            variables = ["lon", "lat", "time"]
            data = read_many_files_individually(filenames, variables)
        except InvalidVariableError:
            variables = ["longitude", "latitude", "time"]
            data = read_many_files_individually(filenames, variables)

        logging.info("Listing coordinates: " + str(variables))

        coords = CoordList()
        coords.append(Coord(data[variables[0]], get_metadata(data[variables[0]][0]), "X"))
        coords.append(Coord(data[variables[1]], get_metadata(data[variables[1]][0]), "Y"))
        coords.append(self._fix_time(Coord(data[variables[2]], get_metadata(data[variables[2]][0]), "T")))

        return coords

    @abstractmethod
    def _fix_time(self, coords):
        pass

    def create_coords(self, filenames, variable=None):
        return UngriddedCoordinates(self._create_coord_list(filenames))

    def create_data_object(self, filenames, variable):
        from cis.data_io.netcdf import get_metadata, read_many_files_individually

        coords = self._create_coord_list(filenames)
        var = read_many_files_individually(filenames, [variable])
        metadata = get_metadata(var[variable][0])

        return UngriddedData(var[variable], metadata, coords)


class Cloud_CCI_L2(CCI_L2, AProduct):
    """Climate Change Initiative cloud data at satellite resolution (1km)."""

    def get_file_signature(self):
        return [r'.*ESACCI-L2_CLOUD.*']

    def _fix_time(self, coord):
        coord.convert_julian_to_std_time()
        return coord

    def get_file_format(self, filenames):
        return "NetCDF/Cloud_CCI_L2"


class Aerosol_CCI_L2(CCI_L2, AProduct):
    """Climate Change Initiative aerosol data at near-satellite resolution (10km)."""

    valid_dimensions = ["pixel_number"]

    def get_file_signature(self):
        return [r'.*ESACCI-L2P_AEROSOL.*']

    def _fix_time(self, coord):
        import datetime
        coord.convert_TAI_time_to_std_time(datetime.datetime(1970, 1, 1))
        return coord

    def get_file_format(self, filename):
        return "NetCDF/Aerosol_CCI_L2"


def convert_time_string_to_number(time_str, units):
    from datetime import datetime
    from dateutil.relativedelta import relativedelta

    if "44T" in time_str:
        # Workaround a bug in the generation of timestamps in monthly Level 3 data
        time = datetime.strptime(time_str.replace("44T", "T"), "%Y%mT%H%M%SZ")
        time += relativedelta(months=1, days=-1)
    else:
        time = datetime.strptime(time_str, "%Y%m%dT%H%M%SZ")
    return units.date2num(time)


class Cloud_CCI_L3(NetCDF_Gridded):
    """Climate Change Initiative cloud data on a rectangular lat/lon grid."""

    def get_file_signature(self):
        return [r'.*ESACCI-L3C_CLOUD.*nc', r'.*ESACCI-L3U_CLOUD.*nc']

    def get_file_format(self, filenames):
        return "NetCDF/Cloud_CCI_L3"

    def load_single_file_callback(self, cube, field, filename):
        """Add time coordinate to cube"""
        time_coord = cube.coord("time")

        # Pop time attributes to remove them from the cube
        time_bounds = [convert_time_string_to_number(cube.attributes.pop(att), time_coord.units)
                       for att in ("time_coverage_start", "time_coverage_end")]
        time_coord.bounds = time_bounds

    def load_multiple_files_callback(self, cube, field, filename):
        """Remove attributes that vary from file to file"""
        for att in ('number_of_processed_orbits', 'id', 'tracking_id', 'date_created'):
            try:
                _ = cube.attributes.pop(att)
            except KeyError:
                pass

        self.load_single_file_callback(cube, field, filename)


class Aerosol_CCI_L3(NetCDF_Gridded):
    """Climate Change Initiative aerosol data on a rectangular lat/lon grid."""

    def get_file_signature(self):
        return [r'.*ESACCI-L3C_AEROSOL.*nc']

    def get_file_format(self, filenames):
        return "NetCDF/Aerosol_CCI_L3"

    def load_single_file_callback(self, cube, field, filename):
        """Add time coordinate"""
        from iris.coords import AuxCoord

        # Pop time attributes to remove them from the cube
        try:
            time_bounds = [convert_time_string_to_number(cube.attributes.pop(att), ESA_TIME_UNIT)
                           for att in ("time_coverage_start", "time_coverage_end")]
        except ValueError:
            if "201101-ESACCI" in filename:
                # Manually fix a bugged file
                # (201101-ESACCI-L3C_AEROSOL-AER_PRODUCTS-AATSR-ENVISAT-ADV_MOTNHLY-v2.30.nc)
                time_bounds = [convert_time_string_to_number(s, time_coord.units)
                               for s in ("20110101T000000Z", "20110131T235959Z")]
            else:
                raise ValueError("Bad time_coverage attribute in {}".format(filename))

        time_centre = 0.5 * sum(time_bounds)
        time_coord = AuxCoord(time_centre, standard_name="time", units=ESA_TIME_UNIT, bounds=time_bounds)
        cube.add_aux_coord(time_coord)

    def load_multiple_files_callback(self, cube, field, filename):
        """Remove attributes that vary from file to file"""
        for att in ('id', 'tracking_id', 'date_created', 'time_coverage_duration', 'history', 'inputfilelist'):
            try:
                _ = cube.attributes.pop(att)
            except KeyError:
                pass

        self.load_single_file_callback(cube, field, filename)



"""
module to test the hdf4 utility function of hdf_sd.py
"""
from nose.tools import istest, eq_

from cis.test.integration_test_data import valid_hdf_sd_file, escape_colons, skip_pyhdf
import cis.data_io.hdf_sd as hdf_sd


@istest
@skip_pyhdf
def test_that_can_read_all_variables():
    data_dict = hdf_sd.read(escape_colons(valid_hdf_sd_file))
    eq_(len(data_dict), 67)


@istest
@skip_pyhdf
def test_that_can_read_known_variables():
    data_dict = hdf_sd.read(escape_colons(valid_hdf_sd_file), ['Latitude', 'Longitude'])
    eq_(len(data_dict), 2)


@istest
@skip_pyhdf
def test_that_can_get_data():
    data_dict = hdf_sd.read(escape_colons(valid_hdf_sd_file))
    data = hdf_sd.get_data(data_dict['Latitude'])
    eq_(data.shape, (203, 135))


@istest
@skip_pyhdf
def test_that_can_get_metadata_for_known_variable():
    data_dict = hdf_sd.read(escape_colons(valid_hdf_sd_file))
    metadata = hdf_sd.get_metadata(data_dict['Latitude'])

    eq_(metadata._name, "Latitude")
    eq_(metadata.standard_name, "latitude")
    eq_(metadata.long_name, "Geodetic Latitude")
    eq_(metadata.shape, [203, 135])
    eq_(metadata.units, "Degrees_north")
    eq_(metadata.factor, 1.0)
    eq_(metadata.offset, 0.0)
    eq_(metadata.missing_value, -999.0)

    attr = metadata.misc
    eq_(len(attr), 5)
    eq_(attr['Parameter_Type'], "MODIS Input")
    eq_(attr['valid_range'], [-90.0, 90.0])



import unittest

from hamcrest import assert_that, is_, ends_with
import numpy

from cis.data_io.gridded_data import make_from_cube, GriddedDataList, GriddedData
from cis.evaluate import Calculator, EvaluationError
from cis.test.util import mock
from cis.test.utils_for_testing import compare_masked_arrays
from cis import __version__


class TestCalculator(unittest.TestCase):
    def setUp(self):
        self.calc = Calculator()
        self.data = GriddedDataList([make_from_cube(mock.make_mock_cube())])
        self.data[0].var_name = 'var_name'

    def _make_two_gridded(self):
        data1 = make_from_cube(mock.make_mock_cube())
        data2 = make_from_cube(mock.make_mock_cube(data_offset=10))
        data1.var_name = 'var1'
        data2.var_name = 'var2'
        data1.filenames = ['filename1']
        data2.filenames = ['filename2']
        self.data = [data1, data2]
        self.data = GriddedDataList([data1, data2])

    def _make_two_ungridded_data(self):
        data1 = mock.make_regular_2d_ungridded_data_with_missing_values()
        data2 = mock.make_regular_2d_ungridded_data_with_missing_values()
        data1.metadata._name = 'var1'
        data2.metadata._name = 'var2'
        data1.filenames = ['filename1']
        data2.filenames = ['filename2']
        self.data = [data1, data2]

    def test_GIVEN_expr_with_double_underscores_WHEN_calculate_THEN_raises_EvaluationError(self):
        expr = "[c for c in ().__class__.__base__.__subclasses__() if c.__name__ " \
               "== 'catch_warnings'][0]()._module.__builtins__['__import__']('os')"
        with self.assertRaises(EvaluationError):
            self.calc.evaluate(self.data, expr)

    def test_GIVEN_expr_using_disallowed_builtins_WHEN_calculate_THEN_raises_EvaluationError(self):
        expr = 'open("path")'
        with self.assertRaises(EvaluationError):
            self.calc.evaluate(self.data, expr)

    def test_GIVEN_expr_using_numpy_WHEN_calculate_THEN_allowed(self):
        expr = 'numpy.log(var_name)'
        self.calc.evaluate(self.data, expr)

    def test_GIVEN_expr_using_allowed_builtins_WHEN_calculate_THEN_allowed(self):
        expr = 'var_name + sum(sum(var_name))'
        self.calc.evaluate(self.data, expr)

    def test_GIVEN_two_cubes_and_basic_addition_WHEN_calculate_THEN_addition_successful(self):
        self._make_two_gridded()
        expr = 'var1 + var2'

        res = self.calc.evaluate(self.data, expr)
        expected = numpy.array([[12, 14, 16], [18, 20, 22], [24, 26, 28], [30, 32, 34], [36, 38, 40]])

        assert_that(numpy.array_equal(res.data, expected))

    def test_GIVEN_two_cubes_interpolated_WHEN_calculate_THEN_interpolation_successful(self):
        self._make_two_gridded()
        # Simulate the use case of interpolating between two wavelengths
        # 550 -> [600] -> 670
        expr = 'var1 + (var2 - var1) * (600 - 550) / (670 - 550)'

        res = self.calc.evaluate(self.data, expr)
        expected = numpy.array([[5, 6, 7], [8, 9, 10], [11, 12, 13], [14, 15, 16], [17, 18, 19]]) + 1.0 / 6

        assert_that(numpy.allclose(res.data, expected))

    def test_GIVEN_ungridded_data_basic_addition_WHEN_calculate_THEN_addition_successful(self):
        data1 = mock.make_regular_2d_ungridded_data()
        data2 = mock.make_regular_2d_ungridded_data()
        data1.metadata._name = 'var1'
        data2.metadata._name = 'var2'
        self.data = [data1, data2]
        expr = 'var1 + var2'

        res = self.calc.evaluate(self.data, expr)
        expected = 2 * self.data[1].data.flatten()

        assert_that(numpy.array_equal(res.data, expected))

    def test_GIVEN_ungridded_data_one_flattened_WHEN_calculate_THEN_compared_on_flattened_grid(self):
        data1 = mock.make_regular_2d_ungridded_data()
        data2 = mock.make_regular_2d_ungridded_data()
        data2._data = data2.data_flattened
        for coord in data2.coords():
            coord._data = coord.data_flattened
        data1.metadata._name = 'var1'
        data2.metadata._name = 'var2'
        self.data = [data1, data2]
        expr = 'var1 + var2'
        res = self.calc.evaluate(self.data, expr)
        expected = 2 * self.data[1].data
        assert_that(numpy.array_equal(res.data, expected))
        assert_that(res.data.shape, is_((15,)))

    def test_GIVEN_ungridded_missing_values_WHEN_calculate_THEN_missing_values_preserved(self):
        data = mock.make_regular_2d_ungridded_data_with_missing_values()
        data.metadata._name = 'var1'
        self.data = [data]
        expr = 'var1 + 10'

        res = self.calc.evaluate(self.data, expr)
        expected = numpy.ma.masked_invalid([[11, 12, 13], [14, float('Nan'), 16], [17, 18, float('Nan')],
                                            [20, 21, 22], [float('Nan'), 24, 25]]).flatten()
        compare_masked_arrays(res.data, expected)

    def test_GIVEN_two_variables_WHEN_calculate_THEN_variable_values_not_changed(self):
        # This is to fix an issue where the calculator was taking a reference to the input list and therefore changing
        # its value when output.
        data1 = mock.make_regular_2d_ungridded_data()
        data2 = mock.make_regular_2d_ungridded_data()
        data1.metadata._name = 'var1'
        data2.metadata._name = 'var2'
        self.data = [data1, data2]
        expr = 'var1 + var2'

        res = self.calc.evaluate(self.data, expr)
        assert_that(numpy.array_equal(data1.data, data2.data))

    def test_GIVEN_two_cubes_basic_addition_WHEN_calculate_THEN_metadata_correct(self):
        self._make_two_gridded()
        expr = 'var1 + alias2'
        self.data[1].alias = 'alias2'

        res = self.calc.evaluate(self.data, expr)
        expected_var_name = 'calculated_variable'
        expected_standard_name = None
        expected_long_name = 'Calculated value for expression "%s"' % expr

        assert_that(isinstance(res, GriddedData))
        assert_that(res.var_name, is_(expected_var_name))
        assert_that(res.standard_name, is_(expected_standard_name))
        assert_that(res.long_name, is_(expected_long_name))
        assert_that(res.alias, is_(res.var_name))

    def test_GIVEN_variables_from_same_file_WHEN_calculate_THEN_history_added(self):
        self._make_two_ungridded_data()
        self.data[1].alias = 'alias2'
        self.data[1].filenames = self.data[0].filenames
        expr = 'var1 + alias2'
        res = self.calc.evaluate(self.data, expr)
        expected_history = "Evaluated using CIS version " + __version__ + \
                           "\nExpression evaluated: 'var1 + alias2'" + \
                           "\nwith variables: 'var1' from files ['filename1']," + \
                           "\n'var2' (as 'alias2') from files ['filename1']."
        # Do an ends_with comparison because history starts with timestamp
        assert_that(res.history, ends_with(expected_history))

    def test_GIVEN_variables_from_different_files_WHEN_calculate_THEN_history_added(self):
        self._make_two_gridded()
        expr = 'var1 + var2'
        res = self.calc.evaluate(self.data, expr)
        expected_history = "Evaluated using CIS version " + __version__ + \
                           "\nExpression evaluated: 'var1 + var2'" + \
                           "\nwith variables: 'var1' from files ['filename1']," + \
                           "\n'var2' from files ['filename2']."
        # Do an ends_with comparison because history starts with timestamp
        assert_that(res.history, ends_with(expected_history))

    def test_GIVEN_output_var_name_WHEN_calculate_THEN_output_uses_var_name(self):
        self._make_two_gridded()
        expr = 'var1 + var2'
        res = self.calc.evaluate(self.data, expr, output_var='var_out_name')
        assert_that(res.var_name, is_('var_out_name'))

    def test_GIVEN_operation_results_in_wrong_shape_THEN_raises_EvaluationError(self):
        data1 = mock.make_regular_2d_ungridded_data()
        data1.metadata._name = 'var1'
        self.data = [data1]
        expr = 'numpy.mean(var1)'
        with self.assertRaises(EvaluationError):
            res = self.calc.evaluate(self.data, expr)

    def test_GIVEN_variables_not_compatible_shape_THEN_raises_EvaluationError(self):
        data1 = mock.make_regular_2d_ungridded_data()
        data2 = mock.make_regular_2d_ungridded_data(lat_dim_length=6)
        data1.metadata._name = 'var1'
        data2.metadata._name = 'var2'
        self.data = [data1, data2]
        expr = 'var1 + var2'
        with self.assertRaises(EvaluationError):
            res = self.calc.evaluate(self.data, expr)

    def test_GIVEN_units_string_WHEN_calculate_ungridded_THEN_units_set_on_output(self):
        self._make_two_ungridded_data()
        expr = 'var1 + var2'
        units = 'kg m^-3'
        res = self.calc.evaluate(self.data, expr, units=units)
        assert_that(res.units, is_(units))

    def test_GIVEN_units_string_WHEN_calculate_gridded_THEN_units_set_on_output(self):
        self._make_two_gridded()
        expr = 'var1 + var2'
        units = 'kg m^-3'
        res = self.calc.evaluate(self.data, expr, units=units)
        assert_that(res.units, is_(units))

    def test_GIVEN_attributes_WHEN_calculate_ungridded_THEN_attributes_set_on_output(self):
        self._make_two_ungridded_data()
        expr = 'var1 + var2'
        attributes = {'att1': 'val1', 'att2': 'val2'}
        res = self.calc.evaluate(self.data, expr, attributes=attributes)
        assert_that(res.attributes['att1'], is_('val1'))
        assert_that(res.attributes['att2'], is_('val2'))

    def test_GIVEN_attributes_WHEN_calculate_gridded_THEN_attributes_set_on_output(self):
        self._make_two_gridded()
        expr = 'var1 + var2'
        attributes = {'att1': 'val1', 'att2': 'val2'}
        res = self.calc.evaluate(self.data, expr, attributes=attributes)
        assert_that(res.attributes['att1'], is_('val1'))
        assert_that(res.attributes['att2'], is_('val2'))



"""
Module to test the one-dimensional plotting of NetCDF files
More tests can be found in the manual_integration_tests package
"""
import unittest

import iris
from iris.coords import DimCoord
from iris.cube import Cube
import numpy as np

from cis.data_io.gridded_data import make_from_cube
from cis.plotting.formatted_plot import Plotter
from cis.plotting.genericplot import Generic2DPlot
from cis.test.utils_for_testing import assert_arrays_equal


class TestPlotting(unittest.TestCase):

    def test_should_raise_io_error_with_invalid_filename(self):
        with self.assertRaises(IOError):
            cube = iris.load_cube("/")
            Plotter([cube], "line", "/")

    def test_units_formatting(self):
        from cis.plotting.plot import format_units
        from cf_units import Unit
        assert format_units("kg s-1") == "(kg s-1)"
        assert format_units("seconds since 1600") == ""  # We don't want any units as they will be converted to DateTime
        assert format_units(Unit("kg s-1")) == "(kg s-1)"

    def test_get_label(self):
        from cis.plotting.plot import get_label
        from cis.test.util.mock import make_dummy_1d_ungridded_data
        d = make_dummy_1d_ungridded_data()
        assert get_label(d) == "TOTAL RAINFALL RATE: LS+CONV KG/M2/S (kg m-2 s-1)"
        assert get_label(d, False) == "TOTAL RAINFALL RATE: LS+CONV KG/M2/S"
        d.var_name = ""
        d.long_name = ""
        assert get_label(d) == "unknown (kg m-2 s-1)"
        assert get_label(d, False) == "unknown"

    def test_get_axis_ungridded(self):
        from cis.plotting.plot import get_axis
        from cis.test.util.mock import make_dummy_2d_ungridded_data
        d = make_dummy_2d_ungridded_data()
        assert get_axis(d, "x").name() == 'latitude'
        assert get_axis(d, "x", 'latitude').name() == 'latitude'
        assert get_axis(d, "x", 'bad_name').name() == 'latitude'  # Falls back on axis name
        assert get_axis(d, "y").name() == 'longitude'
        assert get_axis(d, "y", 'longitude').name() == 'longitude'
        assert get_axis(d, "y", 'bad_name').name() == 'longitude'  # Falls back on axis name

    def test_get_axis_gridded(self):
        from cis.plotting.plot import get_axis
        from cis.test.util.mock import make_mock_cube
        from cis.data_io.gridded_data import make_from_cube
        d = make_from_cube(make_mock_cube())
        assert get_axis(d, "x").name() == 'longitude'
        assert get_axis(d, "y").name() == 'latitude'


class TestHeatMap(unittest.TestCase):
    kwargs = {}

    def test_lat_lon_increasing_no_bounds_over_greenwich(self):
        x = np.array([-0.5, 0.5])
        y = np.array([50.5, 51.5])
        values = np.array([[1, 2], [3, 4]])
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        expected_x = np.array([-1, 0, 1])
        expected_y = np.array([50, 51, 52])
        expected_v = np.array([[1, 2],
                               [3, 4]])
        assert_arrays_equal(out_x, expected_x)
        assert_arrays_equal(out_y, expected_y)
        assert_arrays_equal(out_values, expected_v)

    def test_lat_lon_increasing_no_bounds(self):
        x = np.array([0.5, 1.5])
        y = np.array([50.5, 51.5])
        values = np.array([[1, 2], [3, 4]])
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        expected_x = np.array([0, 1, 2])
        expected_y = np.array([50, 51, 52])
        expected_v = np.array([[1, 2],
                               [3, 4]])
        assert_arrays_equal(out_x, expected_x)
        assert_arrays_equal(out_y, expected_y)
        assert_arrays_equal(out_values, expected_v)

    def test_lat_lon_decreasing_no_bounds(self):
        x = np.array([0.5, -0.5])
        y = np.array([51.5, 50.5])
        values = np.array([[1, 2], [3, 4]])
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        expected_x = np.array([1, 0, -1])
        expected_y = np.array([52, 51, 50])
        expected_v = np.array([[1, 2],
                               [3, 4]])
        assert_arrays_equal(out_x, expected_x)
        assert_arrays_equal(out_y, expected_y)
        assert_arrays_equal(out_values, expected_v)

    def test_wide_longitude(self):
        x = np.arange(-174, 186, 10)
        y = np.array([50.5, 51.5])
        values = np.arange(len(y) * len(x)).reshape((len(y), len(x)))
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        x_bounds = np.arange(-179, 190, 10)
        y_bounds = np.array([50, 51, 52])
        assert_arrays_equal(out_x, x_bounds)
        assert_arrays_equal(out_y, y_bounds)

    def test_longitude_0_360(self):
        x = np.arange(10, 370, 20)
        y = np.array([50.5, 51.5])
        values = np.arange(len(y) * len(x)).reshape((len(y), len(x)))
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        x_bounds = np.arange(0, 380, 20)
        y_bounds = np.array([50, 51, 52])
        assert_arrays_equal(out_x, x_bounds)
        assert_arrays_equal(out_y, y_bounds)

    def test_longitude_0_360_one_degree(self):
        x = np.arange(0.5, 360.5, 1)
        y = np.array([50.5, 51.5])
        values = np.arange(len(y) * len(x)).reshape((len(y), len(x)))
        latitude = DimCoord(y, standard_name='latitude', units='degrees')
        longitude = DimCoord(x, standard_name='longitude', units='degrees')
        data = make_from_cube(Cube(values, dim_coords_and_dims=[(latitude, 0), (longitude, 1)]))
        out_values, out_x, out_y= Generic2DPlot._cube_manipulation(data, longitude.contiguous_bounds(),
                                                                   latitude.contiguous_bounds())
        x_bounds = np.arange(0, 361, 1)
        y_bounds = np.array([50, 51, 52])
        assert_arrays_equal(out_x, x_bounds)
        assert_arrays_equal(out_y, y_bounds)



from cis.data_io.products import NetCDF_Gridded
import logging


class HadGEM_CONVSH(NetCDF_Gridded):
    """
        HadGEM plugin for reading NetCDF files converted by CONVSH. It implements a callback to pass to iris when
        reading multiple files to allow correct merging
    """

    def get_file_signature(self):
        return [r'[a-z]{6}[\._][pamd]{2}[0-9]{4,6}.*\.nc']

    @staticmethod
    def load_multiple_files_callback(cube, field, filename):
        from iris.util import squeeze
        # We need to remove the history field when reading multiple files so that the cubes can be properly merged
        cube.attributes.pop('history')
        # cube.coord(name_or_coord='Hybrid height').attributes['formula_terms'] = 'a: lev b: b orog: orog'
        # We also need to remove the length one time dimension so that the cube can be merged correctly (iris preserves
        #  the value as a scalar which then gets converted back into a full coordinate again on merge).
        return squeeze(cube)

    def _create_cube(self, filenames, variable):
        """Creates a cube for the specified variable.
        :param filenames: List of filenames to read coordinates from
        :param variable: Optional variable to read while we're reading the coordinates, can be a string or a
        VariableConstraint object
        :return: If variable was specified this will return an UngriddedData object, otherwise a CoordList
        """
        import six
        from cis.exceptions import InvalidVariableError
        from cis.data_io.products.gridded_NetCDF import DisplayConstraint
        from cis.data_io.gridded_data import load_cube
        from iris.exceptions import CoordinateNotFoundError

        # Check if the files given actually exist.
        for filename in filenames:
            with open(filename) as f:
                pass

        variable_constraint = variable
        if isinstance(variable, six.string_types):
            # noinspection PyPep8
            variable_constraint = DisplayConstraint(cube_func=(lambda c: c.var_name == variable or
                                                               c.standard_name == variable or
                                                               c.long_name == variable), display=variable)
        if len(filenames) == 1:
            callback_function = self.load_single_file_callback
        else:
            callback_function = self.load_multiple_files_callback

        try:
            cube = load_cube(filenames, variable_constraint, callback=callback_function)
        except ValueError as e:
            if variable is None:
                message = "File contains more than one cube variable name must be specified"
            elif e.args[0] == "No cubes found":
                message = "Variable not found: {} \nTo see a list of variables run: cis info {}" \
                    .format(str(variable), filenames[0])
            else:
                message = e.args[0]
            raise InvalidVariableError(message)

        try:
            hybrid_ht = cube.coord(name_or_coord='Hybrid height')
            hybrid_ht.attributes['formula'] = 'z(k,j,i) = a(k) + b(k)*orog(j,i)'
            hybrid_ht.convert_units('m')
        except CoordinateNotFoundError as e:
            pass

        try:
            cube.coord(long_name='t').standard_name = 'time'
        except CoordinateNotFoundError as e:
            pass

        self._add_available_aux_coords(cube, filenames)

        return cube

    def get_variable_names(self, filenames, data_type=None):
        # Don't do any checks on valid variables at the moment as iris can't parse the hybrid height dimension units...
        import iris
        from cis.utils import single_warnings_only
        # Filter the warnings so that they only appear once - otherwise you get lots of repeated warnings
        with single_warnings_only():
            cubes = iris.load(filenames)

        return set(cube.name() for cube in cubes)


class HadGEM_PP(NetCDF_Gridded):
    """
        HadGEM plugin for reading native pp files
    """

    def get_file_signature(self):
        return [r'.*\.pp']

    @staticmethod
    def load_multiple_files_callback(cube, field, filename):
        # This method sets the var_name (used for outputting the cube to NetCDF) to the cube name. This can be quite
        #  for some HadGEM variables but most commands allow the user to override this field on output.
        var_name = cube.name()
        try:
            cube.var_name = var_name
        except ValueError as e:
            logging.info("Unable to set var_name due to error: {}".format(e))

    @staticmethod
    def load_single_file_callback(cube, field, filename):
        # This method sets the var_name (used for outputting the cube to NetCDF) to the cube name. This can be quite
        #  for some HadGEM variables but most commands allow the user to override this field on output.
        var_name = cube.name()
        try:
            cube.var_name = var_name
        except ValueError as e:
            try:
                cube.var_name = var_name.replace(' ', '_')
            except ValueError as e:
                logging.info("Unable to set var_name due to error: {}".format(e))



# (C) British Crown Copyright 2010 - 2016, Met Office
#
# This file was derived from a related file in the Iris project which is released under the LGPL 3
"""
Basic mathematical and statistical operations.
"""
import math
import operator

import cf_units
import numpy as np


def abs(ungridded_data, in_place=False):
    """
    Calculate the absolute values of the data in the LazyData provided.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    return _math_op_common(ungridded_data, np.abs, ungridded_data.units, in_place=in_place)


def _assert_is_ungridded_data(ungridded_data):
    from cis.data_io.ungridded_data import LazyData
    if not isinstance(ungridded_data, LazyData):
        raise TypeError('The "ungridded_data" argument must be an instance of '
                        'cis.data_io.ungridded_data.LazyData.')


def _assert_matching_units(ungridded_data, other, operation_name):
    """
    Check that the units of the ungridded_data and the other item are the same, or if
    the other does not have a unit, skip this test
    """
    if ungridded_data.units != getattr(other, 'units', ungridded_data.units):
        msg = 'Cannot use {!r} with differing units ({} & {})'.format(
            operation_name, ungridded_data.units, other.units)
        raise NotImplementedError(msg)


def add(ungridded_data, other, in_place=False):
    """
    Calculate the sum of two ungridded_datas, or the sum of a ungridded_data and a
    coordinate or scalar value.

    When summing two ungridded_datas, they must both have the same coordinate
    systems & data resolution.

    When adding a coordinate to a ungridded_data, they must both share the same
    number of elements along a shared axis.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.
    * other:
        An instance of :class:`cis.data_io.ungridded_data.LazyData` 
        or a number or :class:`numpy.ndarray`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    op = operator.iadd if in_place else operator.add
    return _add_subtract_common(op, ungridded_data, other, in_place=in_place)


def subtract(ungridded_data, other, in_place=False):
    """
    Calculate the difference between two ungridded_datas, or the difference between
    a ungridded_data and a coordinate or scalar value.

    When subtracting two ungridded_datas, they must both have the same coordinate
    systems & data resolution.

    When subtracting a coordinate to a ungridded_data, they must both share the
    same number of elements along a shared axis.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.
    * other:
        An instance of :class:`cis.data_io.ungridded_data.LazyData` 
        or a number or :class:`numpy.ndarray`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    op = operator.isub if in_place else operator.sub
    return _add_subtract_common(op, ungridded_data, other, in_place=in_place)


def _add_subtract_common(operation_function, ungridded_data, other, in_place=False):
    """
    Function which shares common code between addition and subtraction
    of ungridded_datas.

    operation_function   - function which does the operation
                           (e.g. numpy.subtract)
    ungridded_data                 - the ungridded_data whose data is used as the first argument
                           to `operation_function`
    other                - the ungridded_data, coord, ndarray or number whose data is
                           used as the second argument
    in_place             - whether or not to apply the operation in place to
                           `ungridded_data` and `ungridded_data.data`

    """
    _assert_is_ungridded_data(ungridded_data)
    _assert_matching_units(ungridded_data, other, operation_function.__name__)

    new_ungridded_data = _binary_op_common(operation_function, ungridded_data, other, ungridded_data.units, in_place)

    return new_ungridded_data


def multiply(ungridded_data, other, in_place=False):
    """
    Calculate the product of a ungridded_data and another ungridded_data or coordinate.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.
    * other:
        An instance of :class:`cis.data_io.ungridded_data.LazyData` 
        or a number or :class:`numpy.ndarray`.

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    _assert_is_ungridded_data(ungridded_data)
    other_unit = getattr(other, 'units', '1')
    new_unit = ungridded_data.units * other_unit
    op = operator.imul if in_place else operator.mul
    return _binary_op_common(op, ungridded_data, other, new_unit, in_place=in_place)


def divide(ungridded_data, other, in_place=False):
    """
    Calculate the division of a ungridded_data by a ungridded_data or coordinate.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.
    * other:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`
        or a number or :class:`numpy.ndarray`.

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    _assert_is_ungridded_data(ungridded_data)
    other_unit = getattr(other, 'units', '1')
    new_unit = ungridded_data.units / other_unit
    try:
        op = operator.idiv if in_place else operator.div
    except AttributeError:
        op = operator.itruediv if in_place else operator.truediv
    return _binary_op_common(op, ungridded_data, other, new_unit, in_place=in_place)


def exponentiate(ungridded_data, exponent, in_place=False):
    """
    Returns the result of the given ungridded_data to the power of a scalar.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.
    * exponent:
        The integer or floating point exponent.

        .. note:: When applied to the ungridded_data's unit, the exponent must
            result in a unit that can be described using only integer
            powers of the basic units.

            e.g. Unit('meter^-2 kilogram second^-1')

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    _assert_is_ungridded_data(ungridded_data)

    def power(data, out=None):
        return np.power(data, exponent, out)

    return _math_op_common(ungridded_data, power, ungridded_data.units ** exponent, in_place=in_place)


def exp(ungridded_data, in_place=False):
    """
    Calculate the exponential (exp(x)) of the ungridded_data.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    .. note::

        Taking an exponential will return a ungridded_data with dimensionless units.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    return _math_op_common(ungridded_data, np.exp, cf_units.Unit('1'), in_place=in_place)


def log(ungridded_data, in_place=False):
    """
    Calculate the natural logarithm (base-e logarithm) of the ungridded_data.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    return _math_op_common(ungridded_data, np.log, ungridded_data.units.log(math.e), in_place=in_place)


def log2(ungridded_data, in_place=False):
    """
    Calculate the base-2 logarithm of the ungridded_data.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    return _math_op_common(ungridded_data, np.log2, ungridded_data.units.log(2), in_place=in_place)


def log10(ungridded_data, in_place=False):
    """
    Calculate the base-10 logarithm of the ungridded_data.

    Args:

    * ungridded_data:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    Kwargs:

    * in_place:
        Whether to create a new LazyData, or alter the given "ungridded_data".

    Returns:
        An instance of :class:`cis.data_io.ungridded_data.LazyData`.

    """
    return _math_op_common(ungridded_data, np.log10, ungridded_data.units.log(10), in_place=in_place)


def _assert_compatible(ungridded_data, other):
    """
    Checks to see if ungridded_data.data and another array can be broadcast to
    the same shape using ``numpy.broadcast_arrays``.

    """
    # This code previously returned broadcasted versions of the cube
    # data and the other array. As numpy.broadcast_arrays does not work
    # with masked arrays (it returns them as ndarrays) operations
    # involving masked arrays would be broken.

    try:
        data_view, other_view = np.broadcast_arrays(ungridded_data.data,
                                                    np.asarray(other))
    except ValueError as err:
        # re-raise
        raise ValueError("The array was not broadcastable to the cube's data "
                         "shape. The error message from numpy when "
                         "broadcasting:\n{}\nThe cube's shape was {} and the "
                         "array's shape was {}".format(err, ungridded_data.shape,
                                                       other.shape))

    if ungridded_data.shape != data_view.shape:
        raise ValueError("The array operation would increase the "
                         "dimensionality of the cube. The new cube's data "
                         "would have had to become: {}".format(
                             data_view.shape))


def _binary_op_common(operation_function, ungridded_data, other, new_unit, in_place=False):
    """
    Function which shares common code between binary operations.

    operation_function   - function which does the operation
                           (e.g. numpy.divide)
    ungridded_data                 - the ungridded_data whose data is used as the first argument
                           to `operation_function`
    other                - the ungridded_data, coord, ndarray or number whose data is
                           used as the second argument
    new_unit             - unit for the resulting quantity
    in_place             - whether or not to apply the operation in place to
                           `ungridded_data` and `ungridded_data.data`
    """
    from cis.data_io.ungridded_data import LazyData
    from iris.cube import Cube
    _assert_is_ungridded_data(ungridded_data)

    if isinstance(other, (LazyData, Cube)):
        other = other.data

    # don't worry about checking for other data types (such as scalars or
    # np.ndarrays) because _assert_compatible validates that they are broadcast
    # compatible with cube.data
    _assert_compatible(ungridded_data, other)

    def unary_func(x):
        ret = operation_function(x, other)
        if ret is NotImplemented:
            # explicitly raise the TypeError, so it gets raised even if, for
            # example, `cis.maths.multiply(ungridded_data, other)` is called
            # directly instead of `ungridded_data * other`
            raise TypeError('cannot %s %r and %r objects' %
                            (operation_function.__name__, type(x).__name__,
                             type(other).__name__))
        return ret
    return _math_op_common(ungridded_data, unary_func, new_unit, in_place)


def _math_op_common(ungridded_data, operation_function, new_unit, in_place=False):
    _assert_is_ungridded_data(ungridded_data)
    if in_place:
        new_ungridded_data = ungridded_data
        try:
            operation_function(new_ungridded_data.data, out=new_ungridded_data.data)
        except TypeError:
            # Non ufunc function
            operation_function(new_ungridded_data.data)
    else:
        new_ungridded_data = ungridded_data.copy(data=operation_function(ungridded_data.data))
    new_ungridded_data.units = new_unit
    new_ungridded_data.add_history('Performed {op} operation'.format(op=operation_function.__name__))
    return new_ungridded_data



import logging
from cis.data_io.Coord import Coord, CoordList
from cis.data_io.products.AProduct import AProduct
from cis.data_io.ungridded_data import UngriddedData, Metadata, UngriddedCoordinates
import cis.utils as utils


class cis(AProduct):
    # If a file matches the CIS product signature as well as another signature (e.g. because we aggregated from another
    # data product) we need to prioritise the CIS data product
    priority = 100

    def get_file_signature(self):
        return [r'.*\.nc']

    def create_coords(self, filenames, usr_variable=None):
        from cis.data_io.netcdf import read_many_files_individually, get_metadata, get_netcdf_file_variables
        from cis.data_io.Coord import Coord, CoordList
        from cis.data_io.ungridded_data import UngriddedCoordinates, UngriddedData
        from cis.exceptions import InvalidVariableError

        # We have to read it once first to find out which variables are in there. We assume the set of coordinates in
        # all the files are the same
        file_variables = get_netcdf_file_variables(filenames[0])

        axis_lookup = {"longitude": "x", 'latitude': 'y', 'altitude': 'z', 'time': 't', 'air_pressure': 'p'}

        coord_variables = [(v, axis_lookup[v]) for v in file_variables if v in axis_lookup]

        # Create a copy to contain all the variables to read
        all_variables = list(coord_variables)
        if usr_variable is not None:
            all_variables.append((usr_variable, ''))

        logging.info("Listing coordinates: " + str(all_variables))

        coords = CoordList()
        var_data = read_many_files_individually(filenames, [v[0] for v in all_variables])
        for name, axis in coord_variables:
            try:
                coords.append(Coord(var_data[name], get_metadata(var_data[name][0]), axis=axis))
            except InvalidVariableError:
                pass

        # Note - We don't need to convert this time coord as it should have been written in our
        #  'standard' time unit

        if usr_variable is None:
            res = UngriddedCoordinates(coords)
        else:
            res = UngriddedData(var_data[usr_variable], get_metadata(var_data[usr_variable][0]), coords)

        return res

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, variable)

    def get_file_format(self, filename):
        return "NetCDF/CIS"

    def get_file_type_error(self, filename):
        """
        Test that the file is of the correct signature
        :param filename: the file name for the file
        :return: list fo errors or None
        """
        from cis.data_io.netcdf import get_netcdf_file_attributes
        atts = get_netcdf_file_attributes(filename)
        errors = None
        try:
            source = atts['source']
        except KeyError as ex:
            errors = ['No source attribute found in {}'.format(filename)]
        else:
            if not source.startswith('CIS'):
                errors = ['Source ({}) does not match CIS in {}'.format(source, filename)]
        return errors


class Aeronet(AProduct):
    def get_file_signature(self):
        return [r'.*\.lev20', r'.*\.ONEILL_lev20', r'.*\.ONEILL_20', r'.*\.lev15',
                r'.*\.ONEILL_lev15', r'.*\.ONEILL_15', r'.*All_Sites_Times.*dat',
                r'.*\.all']

    def _create_coord_list(self, filenames, data=None):
        from cis.data_io.ungridded_data import Metadata
        from cis.data_io.aeronet import load_multiple_aeronet
        from cis.time_util import cis_standard_time_unit as ct

        if data is None:
            data = load_multiple_aeronet(filenames)

        coords = CoordList()
        coords.append(Coord(data['longitude'], Metadata(name="Longitude", shape=(len(data),),
                                                        units="degrees_east", range=(-180, 180))))
        coords.append(Coord(data['latitude'], Metadata(name="Latitude", shape=(len(data),),
                                                       units="degrees_north", range=(-90, 90))))
        coords.append(Coord(data['altitude'], Metadata(name="Altitude", shape=(len(data),), units="meters")))
        coords.append(Coord(data["datetime"], Metadata(name="DateTime", standard_name='time', shape=(len(data),),
                                                       units=ct), "X"))

        return coords

    def create_coords(self, filenames, variable=None):
        return UngriddedCoordinates(self._create_coord_list(filenames))

    def create_data_object(self, filenames, variable):
        from cis.data_io.aeronet import load_multiple_aeronet

        data_obj = load_multiple_aeronet(filenames, [variable])

        coords = self._create_coord_list(filenames, data_obj)

        if variable[-1] == ')':
            # The name is text before any brackets, the units is what's after it (minus the closing bracket)
            name_units = variable.split('(')
            name = name_units[0]
            # For Aeronet we can assume that if there are no units then it is unitless (AOT, Angstrom exponent, etc)
            units = name_units[1][:-1] if len(name_units) > 1 else '1'
        else:
            # SDA variable names include () that don't specify units. Ignoring Exact_Wavelengths.
            name = variable
            units = '1'

        return UngriddedData(data_obj[variable],
                             Metadata(name=name, long_name=variable, shape=(len(data_obj),), units=units,
                                      missing_value=-999.0), coords)

    def get_variable_names(self, filenames, data_type=None):
        from cis.data_io.aeronet import get_aeronet_file_variables

        variables = []
        for filename in filenames:
            file_variables = get_aeronet_file_variables(filename)
            variables.extend(file_variables)
        return set(variables)

    def get_file_format(self, filename):
        from cis.data_io.aeronet import get_aeronet_version

        return "ASCII" + get_aeronet_version(filename)


class ASCII_Hyperpoints(AProduct):
    def get_file_signature(self):
        return [r'.*\.txt']

    def get_variable_names(self, filenames, data_type=None):
        return ['value']

    def create_coords(self, filenames, variable=None):
        from cis.data_io.ungridded_data import Metadata
        from numpy import genfromtxt, NaN
        from cis.exceptions import InvalidVariableError
        from cis.time_util import convert_datetime_to_std_time
        import dateutil.parser as du

        array_list = []

        for filename in filenames:
            try:
                array_list.append(genfromtxt(filename, dtype="f8,f8,f8,O,f8",
                                             names=['latitude', 'longitude', 'altitude', 'time', 'value'],
                                             delimiter=',', missing_values='', usemask=True, invalid_raise=True,
                                             converters={"time": du.parse}))
            except:
                raise IOError('Unable to read file ' + filename)

        data_array = utils.concatenate(array_list)
        n_elements = len(data_array['latitude'])

        coords = CoordList()
        coords.append(Coord(data_array["latitude"],
                            Metadata(standard_name="latitude", shape=(n_elements,), units="degrees_north")))
        coords.append(Coord(data_array["longitude"],
                            Metadata(standard_name="longitude", shape=(n_elements,), units="degrees_east")))
        coords.append(
            Coord(data_array["altitude"], Metadata(standard_name="altitude", shape=(n_elements,), units="meters")))

        time_arr = convert_datetime_to_std_time(data_array["time"])
        time = Coord(time_arr,
                     Metadata(standard_name="time", shape=(n_elements,), units="days since 1600-01-01 00:00:00"))
        coords.append(time)

        if variable:
            try:
                data = UngriddedData(data_array['value'], Metadata(name="value", shape=(n_elements,), units="unknown",
                                                                   missing_value=NaN), coords)
            except:
                InvalidVariableError("Value column does not exist in file " + filenames)
            return data
        else:
            return UngriddedCoordinates(coords)

    def create_data_object(self, filenames, variable):
        return self.create_coords(filenames, True)

    def get_file_format(self, filenames):
        """
        Returns the file format
        """
        return "ASCII/ASCIIHyperpoints"



"""
Module to do integration tests of version. Does not check the version is correct just that the command runs
created without errors.
"""
from unittest import TestCase

from cis.cis_main import parse_and_run_arguments


class TestVersionIntegration(TestCase):

    def test_should_do_scatter_plot_of_file_valid_aerosol_cci_file(self):

        arguments = ['version']
        parse_and_run_arguments(arguments)



"""
Tests the various kernels
"""
import unittest

from nose.tools import eq_
from numpy.testing import assert_almost_equal, assert_equal
import numpy as np

from cis.data_io import gridded_data
from cis.data_io.hyperpoint import HyperPoint, HyperPointList
from cis.data_io.ungridded_data import UngriddedData
from cis.test.util import mock


class TestFullAverage(unittest.TestCase):
    def test_basic_col_in_4d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, moments, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=1.0, lon=1.0, alt=12.0, t=dt.datetime(1984, 8, 29, 8, 34))])

        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), moments())
        means = new_data[0]
        std_dev = new_data[1]
        no_points = new_data[2]

        eq_(means.data[0], 25.5)
        assert_almost_equal(std_dev.data[0], np.sqrt(212.5))
        eq_(no_points.data[0], 50)

    def test_basic_col_in_4d_with_pressure_not_altitude(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, moments, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=1.0, lon=1.0, pres=12.0, t=dt.datetime(1984, 8, 29, 8, 34))])

        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), moments())
        means = new_data[0]
        std_dev = new_data[1]
        no_points = new_data[2]

        eq_(means.data[0], 25.5)
        assert_almost_equal(std_dev.data[0], np.sqrt(212.5))
        eq_(no_points.data[0], 50)


class TestNNHorizontal(unittest.TestCase):
    def test_basic_col_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_horizontal, SepConstraintKdtree

        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=1.0, lon=1.0), HyperPoint(lat=4.0, lon=4.0), HyperPoint(lat=-4.0, lon=-4.0)])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal())[0]
        eq_(new_data.data[0], 8.0)
        eq_(new_data.data[1], 12.0)
        eq_(new_data.data[2], 4.0)

    def test_already_collocated_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_horizontal, SepConstraintKdtree

        ug_data = mock.make_regular_2d_ungridded_data()
        # This point already exists on the cube with value 5 - which shouldn't be a problem
        sample_points = UngriddedData.from_points_array([HyperPoint(0.0, 0.0)])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal())[0]
        eq_(new_data.data[0], 8.0)

    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        """
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_horizontal algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same  you would expect the first point to be chosen. This doesn't seem to always be the case but is
                probably down to floating points errors in the haversine calculation as these test points are pretty
                close together. This test is only really for documenting the behaviour for equidistant points.
        """
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_horizontal, SepConstraintKdtree

        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(2.5, 2.5), HyperPoint(-2.5, 2.5), HyperPoint(2.5, -2.5), HyperPoint(-2.5, -2.5)])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal())[0]
        eq_(new_data.data[0], 11.0)
        eq_(new_data.data[1], 5.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)

    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_horizontal, SepConstraintKdtree

        ug_data = mock.make_regular_2d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(5.5, 5.5), HyperPoint(-5.5, 5.5), HyperPoint(5.5, -5.5), HyperPoint(-5.5, -5.5)])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_horizontal())[0]
        eq_(new_data.data[0], 12.0)
        eq_(new_data.data[1], 6.0)
        eq_(new_data.data[2], 10.0)
        eq_(new_data.data[3], 4.0)


class TestNNTime(unittest.TestCase):
    def test_basic_col_with_incompatible_points_throws_an_AttributeError(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree

        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        # Make sample points with no time dimension specified
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(1.0, 1.0), HyperPoint(4.0, 4.0), HyperPoint(-4.0, -4.0)])
        col = GeneralUngriddedCollocator()
        with self.assertRaises(AttributeError):
            new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]

    def test_basic_col_in_2d_with_time(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=1.0, lon=1.0, t=dt.datetime(1984, 8, 29, 8, 34)))
        sample_points.append(HyperPoint(lat=4.0, lon=4.0, t=dt.datetime(1984, 9, 2, 1, 23)))
        sample_points.append(HyperPoint(lat=-4.0, lon=-4.0, t=dt.datetime(1984, 9, 4, 15, 54)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]
        eq_(new_data.data[0], 3.0)
        eq_(new_data.data[1], 7.0)
        eq_(new_data.data[2], 10.0)

    def test_basic_col_with_time(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree
        import numpy as np

        ug_data = mock.make_MODIS_time_steps()

        ref = np.array([0.0, 1.0, 2.0, 3.0])

        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=149751.369618055))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=149759.378055556, ))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=149766.373969907))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=149776.375995371))
        sample_points = UngriddedData.from_points_array(sample_points)

        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]
        assert (np.equal(new_data.data, ref).all())

    def test_already_collocated_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()

        t0 = dt.datetime(1984, 8, 27)
        for d in range(15):
            sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=t0 + dt.timedelta(days=d)))
        sample_points = UngriddedData.from_points_array(sample_points)

        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]
        assert (np.equal(new_data.data, np.arange(15) + 1.0).all())

    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        """
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_time algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same the first point to be chosen.
        """
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        # Choose a time at midday
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1984, 8, 29, 12))])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]
        eq_(new_data.data[0], 3.0)

    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_time, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_2d_with_time_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1984, 8, 26)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1884, 8, 26)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, t=dt.datetime(1994, 8, 27)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_time())[0]
        eq_(new_data.data[0], 1.0)
        eq_(new_data.data[1], 1.0)
        eq_(new_data.data[2], 15.0)


class TestNNAltitude(unittest.TestCase):
    def test_basic_col_with_incompatible_points_throws_a_TypeError(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_altitude, SepConstraintKdtree

        ug_data = mock.make_regular_4d_ungridded_data()
        # Make sample points with no time dimension specified
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(1.0, 1.0), HyperPoint(4.0, 4.0), HyperPoint(-4.0, -4.0)])
        col = GeneralUngriddedCollocator()
        with self.assertRaises(AttributeError):
            new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_altitude())[0]

    def test_basic_col_in_4d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_altitude, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=1.0, lon=1.0, alt=12.0, t=dt.datetime(1984, 8, 29, 8, 34)))
        sample_points.append(HyperPoint(lat=4.0, lon=4.0, alt=34.0, t=dt.datetime(1984, 9, 2, 1, 23)))
        sample_points.append(HyperPoint(lat=-4.0, lon=-4.0, alt=89.0, t=dt.datetime(1984, 9, 4, 15, 54)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_altitude())[0]
        eq_(new_data.data[0], 6.0)
        eq_(new_data.data[1], 16.0)
        eq_(new_data.data[2], 46.0)

    def test_already_collocated_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_altitude, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, alt=80.0, t=dt.datetime(1984, 9, 4, 15, 54))])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_altitude())[0]
        eq_(new_data.data[0], 41.0)

    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        """
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_time algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same the first point to be chosen.
        """
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_altitude, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        # Choose a time at midday
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, alt=35.0, t=dt.datetime(1984, 8, 29, 12))])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_altitude())[0]
        eq_(new_data.data[0], 16.0)

    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_altitude, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, alt=-12.0, t=dt.datetime(1984, 8, 29, 8, 34)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, alt=91.0, t=dt.datetime(1984, 9, 2, 1, 23)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, alt=890.0, t=dt.datetime(1984, 9, 4, 15, 54)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_altitude())[0]
        eq_(new_data.data[0], 1.0)
        eq_(new_data.data[1], 46.0)
        eq_(new_data.data[2], 46.0)


class TestNNPressure(unittest.TestCase):
    def test_basic_col_with_incompatible_points_throws_a_TypeError(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_pressure, SepConstraintKdtree

        ug_data = mock.make_regular_4d_ungridded_data()
        # Make sample points with no time dimension specified
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(1.0, 1.0), HyperPoint(4.0, 4.0), HyperPoint(-4.0, -4.0)])
        col = GeneralUngriddedCollocator()
        with self.assertRaises(AttributeError):
            new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_pressure())[0]

    def test_basic_col_in_4d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_pressure, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=1.0, lon=1.0, pres=12.0, t=dt.datetime(1984, 8, 29, 8, 34)))
        sample_points.append(HyperPoint(lat=4.0, lon=4.0, pres=34.0, t=dt.datetime(1984, 9, 2, 1, 23)))
        sample_points.append(HyperPoint(lat=-4.0, lon=-4.0, pres=89.0, t=dt.datetime(1984, 9, 4, 15, 54)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_pressure())[0]
        eq_(new_data.data[0], 6.0)
        eq_(new_data.data[1], 16.0)
        eq_(new_data.data[2], 46.0)

    def test_already_collocated_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_pressure, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, pres=80.0, t=dt.datetime(1984, 9, 4, 15, 54))])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_pressure())[0]
        eq_(new_data.data[0], 41.0)

    def test_coordinates_exactly_between_points_in_col_ungridded_to_ungridded_in_2d(self):
        """
            This works out the edge case where the points are exactly in the middle or two or more datapoints.
                The nn_pressure algorithm will start with the first point as the nearest and iterates through the
                points finding any points which are closer than the current closest. If two distances were exactly
                the same the first point to be chosen.
        """
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_pressure, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        # Choose a time at midday
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=0.0, lon=0.0, pres=8, t=dt.datetime(1984, 8, 29, 12))])
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_pressure())[0]
        eq_(new_data.data[0], 1.0)

    def test_coordinates_outside_grid_in_col_ungridded_to_ungridded_in_2d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, nn_pressure, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        sample_points = HyperPointList()
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, pres=0.1, t=dt.datetime(1984, 8, 29, 8, 34)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, pres=91.0, t=dt.datetime(1984, 9, 2, 1, 23)))
        sample_points.append(HyperPoint(lat=0.0, lon=0.0, pres=890.0, t=dt.datetime(1984, 9, 4, 15, 54)))
        sample_points = UngriddedData.from_points_array(sample_points)
        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), nn_pressure())[0]
        eq_(new_data.data[0], 1.0)
        eq_(new_data.data[1], 46.0)
        eq_(new_data.data[2], 46.0)


class TestMean(unittest.TestCase):
    def test_basic_col_in_4d(self):
        from cis.collocation.col_implementations import GeneralUngriddedCollocator, mean, SepConstraintKdtree
        import datetime as dt

        ug_data = mock.make_regular_4d_ungridded_data()
        # Note - This isn't actually used for averaging
        sample_points = UngriddedData.from_points_array(
            [HyperPoint(lat=1.0, lon=1.0, alt=12.0, t=dt.datetime(1984, 8, 29, 8, 34))])

        col = GeneralUngriddedCollocator()
        new_data = col.collocate(sample_points, ug_data, SepConstraintKdtree(), mean())[0]
        eq_(new_data.data[0], 25.5)


if __name__ == '__main__':
    unittest.main()
